import	os
import	unittest
from	pygwarts.tests.hagrid					import MediumSet
from	pygwarts.irma.shelve					import LibraryShelf
from	pygwarts.hagrid.thrivables				import Copse
from	pygwarts.hagrid.sprouts					import fssprout
from	pygwarts.hagrid.planting				import Flourish
from	pygwarts.hagrid.cultivation.registering	import PlantRegister
from	pygwarts.hagrid.cultivation.registering	import PlantRegisterQuerier
from	pygwarts.hagrid.cultivation.registering	import WG
from	pygwarts.hagrid.cultivation.registering	import TG
from	pygwarts.hagrid.cultivation.registering	import LG
from	pygwarts.hagrid.cultivation.registering	import WL
from	pygwarts.hagrid.cultivation.registering	import TL
from	pygwarts.hagrid.cultivation.registering	import LL








class RegisterMultiCases(MediumSet):

	"""
		Triple sources registering and querying
	"""

	@classmethod
	def tearDownClass(cls): cls.clean(cls)
	@classmethod
	def setUpClass(cls):

		super().setUpClass()

		class Forest(Copse):

			@fssprout(cls.MEDIUM_SET_SPROUT_3)
			@fssprout(cls.MEDIUM_SET_SPROUT_2)
			@fssprout(cls.MEDIUM_SET_SPROUT_1)
			@PlantRegister("garden")
			class walk(Flourish):		pass
			class garden(LibraryShelf):	pass

		cls.test_case = Forest()
		cls.test_case.walk()
		cls.prq = PlantRegisterQuerier(cls.test_case.garden)


	def test_PRQ_inner(self):

		self.assertTrue(len(self.test_case.garden), 3)
		self.assertIn(self.MEDIUM_SET_SPROUT_1, self.test_case.garden)
		self.assertIn(self.MEDIUM_SET_SPROUT_2, self.test_case.garden)
		self.assertIn(self.MEDIUM_SET_SPROUT_3, self.test_case.garden)

		self.assertTrue(len(self.test_case.garden()), 3)
		self.assertIn(self.MEDIUM_SET_SPROUT_1, self.test_case.garden())
		self.assertIn(self.MEDIUM_SET_SPROUT_2, self.test_case.garden())
		self.assertIn(self.MEDIUM_SET_SPROUT_3, self.test_case.garden())


	def test_register_roots_query(self):

		root1_WG = self.prq.query(key=WG)
		root1_TG = self.prq.query(key=TG)
		root1_LG = self.prq.query(key=LG)
		root1_WL = self.prq.query(key=WL)
		root1_TL = self.prq.query(key=TL)
		root1_LL = self.prq.query(key=LL)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_TG, int)
		self.assertIsInstance(root1_LG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertIsInstance(root1_TL, int)
		self.assertIsInstance(root1_LL, int)

		root2_WG = self.prq.query(key=WG)
		root2_TG = self.prq.query(key=TG)
		root2_LG = self.prq.query(key=LG)
		root2_WL = self.prq.query(key=WL)
		root2_TL = self.prq.query(key=TL)
		root2_LL = self.prq.query(key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_TG, int)
		self.assertIsInstance(root2_LG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertIsInstance(root2_TL, int)
		self.assertIsInstance(root2_LL, int)

		root3_WG = self.prq.query(key=WG)
		root3_TG = self.prq.query(key=TG)
		root3_LG = self.prq.query(key=LG)
		root3_WL = self.prq.query(key=WL)
		root3_TL = self.prq.query(key=TL)
		root3_LL = self.prq.query(key=LL)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_TG, int)
		self.assertIsInstance(root3_LG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertIsInstance(root3_TL, int)
		self.assertIsInstance(root3_LL, int)

		# For multiple sprout the very first sprout walked
		# must be chosen by default, so every call
		# must produce the same result
		self.assertTrue(root1_WG == root2_WG == root3_WG)
		self.assertTrue(root1_TG == root2_TG == root3_TG)
		self.assertTrue(root1_LG == root2_LG == root3_LG)
		self.assertTrue(root1_WL == root2_WL == root3_WL)
		self.assertTrue(root1_TL == root2_TL == root3_TL)
		self.assertTrue(root1_LL == root2_LL == root3_LL)


	def test_branch_register_roots_query(self):

		root1_WG = self.prq.query(self.MEDIUM_SET_SPROUT_1, key=WG)
		root1_TG = self.prq.query(self.MEDIUM_SET_SPROUT_1, key=TG)
		root1_LG = self.prq.query(self.MEDIUM_SET_SPROUT_1, key=LG)
		root1_WL = self.prq.query(self.MEDIUM_SET_SPROUT_1, key=WL)
		root1_TL = self.prq.query(self.MEDIUM_SET_SPROUT_1, key=TL)
		root1_LL = self.prq.query(self.MEDIUM_SET_SPROUT_1, key=LL)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertEqual(root1_TG, 11)
		self.assertEqual(root1_LG, 6)
		self.assertEqual(root1_TL, 2)
		self.assertEqual(root1_LL, 0)

		root2_WG = self.prq.query(self.MEDIUM_SET_SPROUT_2, key=WG)
		root2_TG = self.prq.query(self.MEDIUM_SET_SPROUT_2, key=TG)
		root2_LG = self.prq.query(self.MEDIUM_SET_SPROUT_2, key=LG)
		root2_WL = self.prq.query(self.MEDIUM_SET_SPROUT_2, key=WL)
		root2_TL = self.prq.query(self.MEDIUM_SET_SPROUT_2, key=TL)
		root2_LL = self.prq.query(self.MEDIUM_SET_SPROUT_2, key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 1)
		self.assertEqual(root2_LG, 10)
		self.assertEqual(root2_TL, 1)
		self.assertEqual(root2_LL, 2)

		root3_WG = self.prq.query(self.MEDIUM_SET_SPROUT_3, key=WG)
		root3_TG = self.prq.query(self.MEDIUM_SET_SPROUT_3, key=TG)
		root3_LG = self.prq.query(self.MEDIUM_SET_SPROUT_3, key=LG)
		root3_WL = self.prq.query(self.MEDIUM_SET_SPROUT_3, key=WL)
		root3_TL = self.prq.query(self.MEDIUM_SET_SPROUT_3, key=TL)
		root3_LL = self.prq.query(self.MEDIUM_SET_SPROUT_3, key=LL)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertEqual(root3_TG, 5)
		self.assertEqual(root3_LG, 10)
		self.assertEqual(root3_TL, 1)
		self.assertEqual(root3_LL, 0)


	def test_branch_sprout_register_roots_query(self):

		root1_WG = self.prq.query(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, key=WG)
		root1_TG = self.prq.query(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, key=TG)
		root1_LG = self.prq.query(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, key=LG)
		root1_WL = self.prq.query(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, key=WL)
		root1_TL = self.prq.query(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, key=TL)
		root1_LL = self.prq.query(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, key=LL)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertEqual(root1_TG, 11)
		self.assertEqual(root1_LG, 6)
		self.assertEqual(root1_TL, 2)
		self.assertEqual(root1_LL, 0)

		root2_WG = self.prq.query(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, key=WG)
		root2_TG = self.prq.query(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, key=TG)
		root2_LG = self.prq.query(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, key=LG)
		root2_WL = self.prq.query(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, key=WL)
		root2_TL = self.prq.query(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, key=TL)
		root2_LL = self.prq.query(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 1)
		self.assertEqual(root2_LG, 10)
		self.assertEqual(root2_TL, 1)
		self.assertEqual(root2_LL, 2)

		root3_WG = self.prq.query(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, key=WG)
		root3_TG = self.prq.query(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, key=TG)
		root3_LG = self.prq.query(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, key=LG)
		root3_WL = self.prq.query(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, key=WL)
		root3_TL = self.prq.query(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, key=TL)
		root3_LL = self.prq.query(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, key=LL)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertEqual(root3_TG, 5)
		self.assertEqual(root3_LG, 10)
		self.assertEqual(root3_TL, 1)
		self.assertEqual(root3_LL, 0)


	def test_sprout_register_roots_query(self):

		root1_WG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_1, key=WG)
		root1_TG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_1, key=TG)
		root1_LG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_1, key=LG)
		root1_WL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_1, key=WL)
		root1_TL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_1, key=TL)
		root1_LL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_1, key=LL)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertEqual(root1_TG, 11)
		self.assertEqual(root1_LG, 6)
		self.assertEqual(root1_TL, 2)
		self.assertEqual(root1_LL, 0)

		root2_WG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_2, key=WG)
		root2_TG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_2, key=TG)
		root2_LG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_2, key=LG)
		root2_WL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_2, key=WL)
		root2_TL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_2, key=TL)
		root2_LL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_2, key=LL)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 1)
		self.assertEqual(root2_LG, 10)
		self.assertEqual(root2_TL, 1)
		self.assertEqual(root2_LL, 2)

		root3_WG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_3, key=WG)
		root3_TG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_3, key=TG)
		root3_LG = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_3, key=LG)
		root3_WL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_3, key=WL)
		root3_TL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_3, key=TL)
		root3_LL = self.prq.query(sprout=self.MEDIUM_SET_SPROUT_3, key=LL)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertEqual(root3_TG, 5)
		self.assertEqual(root3_LG, 10)
		self.assertEqual(root3_TL, 1)
		self.assertEqual(root3_LL, 0)








	def test_branch_register_children_query(self):

		child1_WG = self.prq.query(self.d_mss1_c, key=WG)
		child1_TG = self.prq.query(self.d_mss1_c, key=TG)
		child1_LG = self.prq.query(self.d_mss1_c, key=LG)
		child1_WL = self.prq.query(self.d_mss1_c, key=WL)
		child1_TL = self.prq.query(self.d_mss1_c, key=TL)
		child1_LL = self.prq.query(self.d_mss1_c, key=LL)

		self.assertIsInstance(child1_WG, int)
		self.assertIsInstance(child1_WL, int)
		self.assertEqual(child1_TG, 8)
		self.assertEqual(child1_LG, 0)
		self.assertEqual(child1_TL, 8)
		self.assertEqual(child1_LL, 0)

		child2_WG = self.prq.query(self.d_mss2_c, key=WG)
		child2_TG = self.prq.query(self.d_mss2_c, key=TG)
		child2_LG = self.prq.query(self.d_mss2_c, key=LG)
		child2_WL = self.prq.query(self.d_mss2_c, key=WL)
		child2_TL = self.prq.query(self.d_mss2_c, key=TL)
		child2_LL = self.prq.query(self.d_mss2_c, key=LL)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 0)
		self.assertEqual(child2_LG, 8)
		self.assertEqual(child2_TL, 0)
		self.assertEqual(child2_LL, 8)

		child3_WG = self.prq.query(self.d_mss3_k_t, key=WG)
		child3_TG = self.prq.query(self.d_mss3_k_t, key=TG)
		child3_LG = self.prq.query(self.d_mss3_k_t, key=LG)
		child3_WL = self.prq.query(self.d_mss3_k_t, key=WL)
		child3_TL = self.prq.query(self.d_mss3_k_t, key=TL)
		child3_LL = self.prq.query(self.d_mss3_k_t, key=LL)

		self.assertIsInstance(child3_WG, int)
		self.assertIsInstance(child3_WL, int)
		self.assertEqual(child3_TG, 1)
		self.assertEqual(child3_LG, 5)
		self.assertEqual(child3_TL, 1)
		self.assertEqual(child3_LL, 0)


	def test_branch_sprout_register_children_query(self):

		child1_WG = self.prq.query(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, key=WG)
		child1_TG = self.prq.query(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, key=TG)
		child1_LG = self.prq.query(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, key=LG)
		child1_WL = self.prq.query(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, key=WL)
		child1_TL = self.prq.query(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, key=TL)
		child1_LL = self.prq.query(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, key=LL)

		self.assertIsInstance(child1_WG, int)
		self.assertIsInstance(child1_WL, int)
		self.assertEqual(child1_TG, 8)
		self.assertEqual(child1_LG, 0)
		self.assertEqual(child1_TL, 8)
		self.assertEqual(child1_LL, 0)

		child2_WG = self.prq.query(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, key=WG)
		child2_TG = self.prq.query(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, key=TG)
		child2_LG = self.prq.query(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, key=LG)
		child2_WL = self.prq.query(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, key=WL)
		child2_TL = self.prq.query(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, key=TL)
		child2_LL = self.prq.query(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, key=LL)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 0)
		self.assertEqual(child2_LG, 8)
		self.assertEqual(child2_TL, 0)
		self.assertEqual(child2_LL, 8)

		child3_WG = self.prq.query(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, key=WG)
		child3_TG = self.prq.query(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, key=TG)
		child3_LG = self.prq.query(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, key=LG)
		child3_WL = self.prq.query(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, key=WL)
		child3_TL = self.prq.query(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, key=TL)
		child3_LL = self.prq.query(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, key=LL)

		self.assertIsInstance(child3_WG, int)
		self.assertIsInstance(child3_WL, int)
		self.assertEqual(child3_TG, 1)
		self.assertEqual(child3_LG, 5)
		self.assertEqual(child3_TL, 1)
		self.assertEqual(child3_LL, 0)








	def test_register_roots_stats(self):

		root1_WG = self.prq.WG()
		root1_TG = self.prq.TG()
		root1_LG = self.prq.LG()
		root1_WL = self.prq.WL()
		root1_TL = self.prq.TL()
		root1_LL = self.prq.LL()

		root1_WG_apparent = self.prq.WG(apparent=True)
		root1_WL_apparent = self.prq.WL(apparent=True)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_TG, int)
		self.assertIsInstance(root1_LG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertIsInstance(root1_TL, int)
		self.assertIsInstance(root1_LL, int)

		self.assertIsInstance(root1_WG_apparent, int)
		self.assertIsInstance(root1_WL_apparent, int)

		self.assertLess(root1_WG, root1_WG_apparent)
		self.assertLess(root1_WL, root1_WL_apparent)

		root2_WG = self.prq.WG()
		root2_TG = self.prq.TG()
		root2_LG = self.prq.LG()
		root2_WL = self.prq.WL()
		root2_TL = self.prq.TL()
		root2_LL = self.prq.LL()

		root2_WG_apparent = self.prq.WG(apparent=True)
		root2_WL_apparent = self.prq.WL(apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_TG, int)
		self.assertIsInstance(root2_LG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertIsInstance(root2_TL, int)
		self.assertIsInstance(root2_LL, int)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertLess(root2_WG, root2_WG_apparent)
		self.assertLess(root2_WL, root2_WL_apparent)

		root3_WG = self.prq.WG()
		root3_TG = self.prq.TG()
		root3_LG = self.prq.LG()
		root3_WL = self.prq.WL()
		root3_TL = self.prq.TL()
		root3_LL = self.prq.LL()

		root3_WG_apparent = self.prq.WG(apparent=True)
		root3_WL_apparent = self.prq.WL(apparent=True)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_TG, int)
		self.assertIsInstance(root3_LG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertIsInstance(root3_TL, int)
		self.assertIsInstance(root3_LL, int)

		self.assertIsInstance(root3_WG_apparent, int)
		self.assertIsInstance(root3_WL_apparent, int)

		self.assertLess(root3_WG, root3_WG_apparent)
		self.assertLess(root3_WL, root3_WL_apparent)

		# For multiple sprout the very first sprout walked
		# must be chosen by default, so every call
		# must produce the same result
		self.assertTrue(root1_WG == root2_WG == root3_WG)
		self.assertTrue(root1_TG == root2_TG == root3_TG)
		self.assertTrue(root1_LG == root2_LG == root3_LG)
		self.assertTrue(root1_WL == root2_WL == root3_WL)
		self.assertTrue(root1_TL == root2_TL == root3_TL)
		self.assertTrue(root1_LL == root2_LL == root3_LL)

		self.assertTrue(root1_WG_apparent == root2_WG_apparent == root3_WG_apparent)
		self.assertTrue(root1_WL_apparent == root2_WL_apparent == root3_WL_apparent)


	def test_branch_register_roots_stats(self):

		root1_WG = self.prq.WG(self.MEDIUM_SET_SPROUT_1)
		root1_TG = self.prq.TG(self.MEDIUM_SET_SPROUT_1)
		root1_LG = self.prq.LG(self.MEDIUM_SET_SPROUT_1)
		root1_WL = self.prq.WL(self.MEDIUM_SET_SPROUT_1)
		root1_TL = self.prq.TL(self.MEDIUM_SET_SPROUT_1)
		root1_LL = self.prq.LL(self.MEDIUM_SET_SPROUT_1)

		root1_WG_apparent = self.prq.WG(self.MEDIUM_SET_SPROUT_1, apparent=True)
		root1_WL_apparent = self.prq.WL(self.MEDIUM_SET_SPROUT_1, apparent=True)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertEqual(root1_TG, 11)
		self.assertEqual(root1_LG, 6)
		self.assertEqual(root1_TL, 2)
		self.assertEqual(root1_LL, 0)

		self.assertIsInstance(root1_WG_apparent, int)
		self.assertIsInstance(root1_WL_apparent, int)

		self.assertEqual(root1_WG_apparent - root1_WG, 11 *4096)
		self.assertEqual(root1_WL_apparent - root1_WL, 2 *4096)

		root2_WG = self.prq.WG(self.MEDIUM_SET_SPROUT_2)
		root2_TG = self.prq.TG(self.MEDIUM_SET_SPROUT_2)
		root2_LG = self.prq.LG(self.MEDIUM_SET_SPROUT_2)
		root2_WL = self.prq.WL(self.MEDIUM_SET_SPROUT_2)
		root2_TL = self.prq.TL(self.MEDIUM_SET_SPROUT_2)
		root2_LL = self.prq.LL(self.MEDIUM_SET_SPROUT_2)

		root2_WG_apparent = self.prq.WG(self.MEDIUM_SET_SPROUT_2, apparent=True)
		root2_WL_apparent = self.prq.WL(self.MEDIUM_SET_SPROUT_2, apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 1)
		self.assertEqual(root2_LG, 10)
		self.assertEqual(root2_TL, 1)
		self.assertEqual(root2_LL, 2)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 4096)

		root3_WG = self.prq.WG(self.MEDIUM_SET_SPROUT_3)
		root3_TG = self.prq.TG(self.MEDIUM_SET_SPROUT_3)
		root3_LG = self.prq.LG(self.MEDIUM_SET_SPROUT_3)
		root3_WL = self.prq.WL(self.MEDIUM_SET_SPROUT_3)
		root3_TL = self.prq.TL(self.MEDIUM_SET_SPROUT_3)
		root3_LL = self.prq.LL(self.MEDIUM_SET_SPROUT_3)

		root3_WG_apparent = self.prq.WG(self.MEDIUM_SET_SPROUT_3, apparent=True)
		root3_WL_apparent = self.prq.WL(self.MEDIUM_SET_SPROUT_3, apparent=True)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertEqual(root3_TG, 5)
		self.assertEqual(root3_LG, 10)
		self.assertEqual(root3_TL, 1)
		self.assertEqual(root3_LL, 0)

		self.assertIsInstance(root3_WG_apparent, int)
		self.assertIsInstance(root3_WL_apparent, int)

		self.assertEqual(root3_WG_apparent - root3_WG, 5 *4096)
		self.assertEqual(root3_WL_apparent - root3_WL, 4096)


	def test_branch_sprout_register_roots_stats(self):

		root1_WG = self.prq.WG(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1)
		root1_TG = self.prq.TG(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1)
		root1_LG = self.prq.LG(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1)
		root1_WL = self.prq.WL(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1)
		root1_TL = self.prq.TL(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1)
		root1_LL = self.prq.LL(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1)

		root1_WG_apparent = self.prq.WG(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, apparent=True)
		root1_WL_apparent = self.prq.WL(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, apparent=True)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertEqual(root1_TG, 11)
		self.assertEqual(root1_LG, 6)
		self.assertEqual(root1_TL, 2)
		self.assertEqual(root1_LL, 0)

		self.assertIsInstance(root1_WG_apparent, int)
		self.assertIsInstance(root1_WL_apparent, int)

		self.assertEqual(root1_WG_apparent - root1_WG, 11 *4096)
		self.assertEqual(root1_WL_apparent - root1_WL, 2 *4096)

		root2_WG = self.prq.WG(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2)
		root2_TG = self.prq.TG(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2)
		root2_LG = self.prq.LG(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2)
		root2_WL = self.prq.WL(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2)
		root2_TL = self.prq.TL(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2)
		root2_LL = self.prq.LL(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2)

		root2_WG_apparent = self.prq.WG(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, apparent=True)
		root2_WL_apparent = self.prq.WL(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 1)
		self.assertEqual(root2_LG, 10)
		self.assertEqual(root2_TL, 1)
		self.assertEqual(root2_LL, 2)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 4096)

		root3_WG = self.prq.WG(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3)
		root3_TG = self.prq.TG(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3)
		root3_LG = self.prq.LG(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3)
		root3_WL = self.prq.WL(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3)
		root3_TL = self.prq.TL(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3)
		root3_LL = self.prq.LL(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3)

		root3_WG_apparent = self.prq.WG(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, apparent=True)
		root3_WL_apparent = self.prq.WL(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, apparent=True)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertEqual(root3_TG, 5)
		self.assertEqual(root3_LG, 10)
		self.assertEqual(root3_TL, 1)
		self.assertEqual(root3_LL, 0)

		self.assertIsInstance(root3_WG_apparent, int)
		self.assertIsInstance(root3_WL_apparent, int)

		self.assertEqual(root3_WG_apparent - root3_WG, 5 *4096)
		self.assertEqual(root3_WL_apparent - root3_WL, 4096)


	def test_sprout_register_roots_stats(self):

		root1_WG = self.prq.WG(sprout=self.MEDIUM_SET_SPROUT_1)
		root1_TG = self.prq.TG(sprout=self.MEDIUM_SET_SPROUT_1)
		root1_LG = self.prq.LG(sprout=self.MEDIUM_SET_SPROUT_1)
		root1_WL = self.prq.WL(sprout=self.MEDIUM_SET_SPROUT_1)
		root1_TL = self.prq.TL(sprout=self.MEDIUM_SET_SPROUT_1)
		root1_LL = self.prq.LL(sprout=self.MEDIUM_SET_SPROUT_1)

		root1_WG_apparent = self.prq.WG(sprout=self.MEDIUM_SET_SPROUT_1, apparent=True)
		root1_WL_apparent = self.prq.WL(sprout=self.MEDIUM_SET_SPROUT_1, apparent=True)

		self.assertIsInstance(root1_WG, int)
		self.assertIsInstance(root1_WL, int)
		self.assertEqual(root1_TG, 11)
		self.assertEqual(root1_LG, 6)
		self.assertEqual(root1_TL, 2)
		self.assertEqual(root1_LL, 0)

		self.assertIsInstance(root1_WG_apparent, int)
		self.assertIsInstance(root1_WL_apparent, int)

		self.assertEqual(root1_WG_apparent - root1_WG, 11 *4096)
		self.assertEqual(root1_WL_apparent - root1_WL, 2 *4096)

		root2_WG = self.prq.WG(sprout=self.MEDIUM_SET_SPROUT_2)
		root2_TG = self.prq.TG(sprout=self.MEDIUM_SET_SPROUT_2)
		root2_LG = self.prq.LG(sprout=self.MEDIUM_SET_SPROUT_2)
		root2_WL = self.prq.WL(sprout=self.MEDIUM_SET_SPROUT_2)
		root2_TL = self.prq.TL(sprout=self.MEDIUM_SET_SPROUT_2)
		root2_LL = self.prq.LL(sprout=self.MEDIUM_SET_SPROUT_2)

		root2_WG_apparent = self.prq.WG(sprout=self.MEDIUM_SET_SPROUT_2, apparent=True)
		root2_WL_apparent = self.prq.WL(sprout=self.MEDIUM_SET_SPROUT_2, apparent=True)

		self.assertIsInstance(root2_WG, int)
		self.assertIsInstance(root2_WL, int)
		self.assertEqual(root2_TG, 1)
		self.assertEqual(root2_LG, 10)
		self.assertEqual(root2_TL, 1)
		self.assertEqual(root2_LL, 2)

		self.assertIsInstance(root2_WG_apparent, int)
		self.assertIsInstance(root2_WL_apparent, int)

		self.assertEqual(root2_WG_apparent - root2_WG, 4096)
		self.assertEqual(root2_WL_apparent - root2_WL, 4096)

		root3_WG = self.prq.WG(sprout=self.MEDIUM_SET_SPROUT_3)
		root3_TG = self.prq.TG(sprout=self.MEDIUM_SET_SPROUT_3)
		root3_LG = self.prq.LG(sprout=self.MEDIUM_SET_SPROUT_3)
		root3_WL = self.prq.WL(sprout=self.MEDIUM_SET_SPROUT_3)
		root3_TL = self.prq.TL(sprout=self.MEDIUM_SET_SPROUT_3)
		root3_LL = self.prq.LL(sprout=self.MEDIUM_SET_SPROUT_3)

		root3_WG_apparent = self.prq.WG(sprout=self.MEDIUM_SET_SPROUT_3, apparent=True)
		root3_WL_apparent = self.prq.WL(sprout=self.MEDIUM_SET_SPROUT_3, apparent=True)

		self.assertIsInstance(root3_WG, int)
		self.assertIsInstance(root3_WL, int)
		self.assertEqual(root3_TG, 5)
		self.assertEqual(root3_LG, 10)
		self.assertEqual(root3_TL, 1)
		self.assertEqual(root3_LL, 0)

		self.assertIsInstance(root3_WG_apparent, int)
		self.assertIsInstance(root3_WL_apparent, int)

		self.assertEqual(root3_WG_apparent - root3_WG, 5 *4096)
		self.assertEqual(root3_WL_apparent - root3_WL, 4096)








	def test_branch_register_children_stats(self):

		child1_WG = self.prq.WG(self.d_mss1_c)
		child1_TG = self.prq.TG(self.d_mss1_c)
		child1_LG = self.prq.LG(self.d_mss1_c)
		child1_WL = self.prq.WL(self.d_mss1_c)
		child1_TL = self.prq.TL(self.d_mss1_c)
		child1_LL = self.prq.LL(self.d_mss1_c)

		child1_WG_apparent = self.prq.WG(self.d_mss1_c, apparent=True)
		child1_WL_apparent = self.prq.WL(self.d_mss1_c, apparent=True)

		self.assertIsInstance(child1_WG, int)
		self.assertIsInstance(child1_WL, int)
		self.assertEqual(child1_TG, 8)
		self.assertEqual(child1_LG, 0)
		self.assertEqual(child1_TL, 8)
		self.assertEqual(child1_LL, 0)

		self.assertIsInstance(child1_WG_apparent, int)
		self.assertIsInstance(child1_WL_apparent, int)

		self.assertEqual(child1_WG_apparent - child1_WG, 8 *4096)
		self.assertEqual(child1_WL_apparent - child1_WL, 8 *4096)

		child2_WG = self.prq.WG(self.d_mss2_c)
		child2_TG = self.prq.TG(self.d_mss2_c)
		child2_LG = self.prq.LG(self.d_mss2_c)
		child2_WL = self.prq.WL(self.d_mss2_c)
		child2_TL = self.prq.TL(self.d_mss2_c)
		child2_LL = self.prq.LL(self.d_mss2_c)

		child2_WG_apparent = self.prq.WG(self.d_mss2_c, apparent=True)
		child2_WL_apparent = self.prq.WL(self.d_mss2_c, apparent=True)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 0)
		self.assertEqual(child2_LG, 8)
		self.assertEqual(child2_TL, 0)
		self.assertEqual(child2_LL, 8)

		self.assertIsInstance(child2_WG_apparent, int)
		self.assertIsInstance(child2_WL_apparent, int)

		self.assertEqual(child2_WG_apparent, child2_WG)
		self.assertEqual(child2_WL_apparent, child2_WL)

		child3_WG = self.prq.WG(self.d_mss3_k_t)
		child3_TG = self.prq.TG(self.d_mss3_k_t)
		child3_LG = self.prq.LG(self.d_mss3_k_t)
		child3_WL = self.prq.WL(self.d_mss3_k_t)
		child3_TL = self.prq.TL(self.d_mss3_k_t)
		child3_LL = self.prq.LL(self.d_mss3_k_t)

		child3_WG_apparent = self.prq.WG(self.d_mss3_k_t, apparent=True)
		child3_WL_apparent = self.prq.WL(self.d_mss3_k_t, apparent=True)

		self.assertIsInstance(child3_WG, int)
		self.assertIsInstance(child3_WL, int)
		self.assertEqual(child3_TG, 1)
		self.assertEqual(child3_LG, 5)
		self.assertEqual(child3_TL, 1)
		self.assertEqual(child3_LL, 0)

		self.assertIsInstance(child3_WG_apparent, int)
		self.assertIsInstance(child3_WL_apparent, int)

		self.assertEqual(child3_WG_apparent - child3_WG, 4096)
		self.assertEqual(child3_WL_apparent - child3_WL, 4096)


	def test_branch_sprout_register_children_stats(self):

		child1_WG = self.prq.WG(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1)
		child1_TG = self.prq.TG(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1)
		child1_LG = self.prq.LG(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1)
		child1_WL = self.prq.WL(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1)
		child1_TL = self.prq.TL(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1)
		child1_LL = self.prq.LL(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1)

		child1_WG_apparent = self.prq.WG(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, apparent=True)
		child1_WL_apparent = self.prq.WL(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, apparent=True)

		self.assertIsInstance(child1_WG, int)
		self.assertIsInstance(child1_WL, int)
		self.assertEqual(child1_TG, 8)
		self.assertEqual(child1_LG, 0)
		self.assertEqual(child1_TL, 8)
		self.assertEqual(child1_LL, 0)

		self.assertIsInstance(child1_WG_apparent, int)
		self.assertIsInstance(child1_WL_apparent, int)

		self.assertEqual(child1_WG_apparent - child1_WG, 8 *4096)
		self.assertEqual(child1_WL_apparent - child1_WL, 8 *4096)

		child2_WG = self.prq.WG(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2)
		child2_TG = self.prq.TG(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2)
		child2_LG = self.prq.LG(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2)
		child2_WL = self.prq.WL(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2)
		child2_TL = self.prq.TL(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2)
		child2_LL = self.prq.LL(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2)

		child2_WG_apparent = self.prq.WG(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, apparent=True)
		child2_WL_apparent = self.prq.WL(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, apparent=True)

		self.assertIsInstance(child2_WG, int)
		self.assertIsInstance(child2_WL, int)
		self.assertEqual(child2_TG, 0)
		self.assertEqual(child2_LG, 8)
		self.assertEqual(child2_TL, 0)
		self.assertEqual(child2_LL, 8)

		self.assertIsInstance(child2_WG_apparent, int)
		self.assertIsInstance(child2_WL_apparent, int)

		self.assertEqual(child2_WG_apparent, child2_WG)
		self.assertEqual(child2_WL_apparent, child2_WL)

		child3_WG = self.prq.WG(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3)
		child3_TG = self.prq.TG(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3)
		child3_LG = self.prq.LG(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3)
		child3_WL = self.prq.WL(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3)
		child3_TL = self.prq.TL(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3)
		child3_LL = self.prq.LL(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3)

		child3_WG_apparent = self.prq.WG(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, apparent=True)
		child3_WL_apparent = self.prq.WL(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, apparent=True)

		self.assertIsInstance(child3_WG, int)
		self.assertIsInstance(child3_WL, int)
		self.assertEqual(child3_TG, 1)
		self.assertEqual(child3_LG, 5)
		self.assertEqual(child3_TL, 1)
		self.assertEqual(child3_LL, 0)

		self.assertIsInstance(child3_WG_apparent, int)
		self.assertIsInstance(child3_WL_apparent, int)

		self.assertEqual(child3_WG_apparent - child3_WG, 4096)
		self.assertEqual(child3_WL_apparent - child3_WL, 4096)








	def test_branch_sprout_content_roots(self):

		for content in (

			self.prq.content(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1),
			self.prq.content(self.MEDIUM_SET_SPROUT_1, self.MEDIUM_SET_SPROUT_1, apparent=True),
		):
			self.assertEqual(len(content),2)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)

		for content in (

			self.prq.content(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2),
			self.prq.content(self.MEDIUM_SET_SPROUT_2, self.MEDIUM_SET_SPROUT_2, apparent=True),
		):
			self.assertEqual(len(content),2)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)

		for content in (

			self.prq.content(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3),
			self.prq.content(self.MEDIUM_SET_SPROUT_3, self.MEDIUM_SET_SPROUT_3, apparent=True),
		):
			self.assertEqual(len(content),1)
			self.assertIsInstance(content[0],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)








	def test_branch_sprout_content_children(self):

		for content in (

			self.prq.content(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1),
			self.prq.content(self.d_mss1_c, self.MEDIUM_SET_SPROUT_1, apparent=True),
		):
			self.assertEqual(len(content),8)
			self.assertIsInstance(content[0],tuple)
			self.assertIsInstance(content[1],tuple)
			self.assertIsInstance(content[2],tuple)
			self.assertIsInstance(content[3],tuple)
			self.assertIsInstance(content[4],tuple)
			self.assertIsInstance(content[5],tuple)
			self.assertIsInstance(content[6],tuple)
			self.assertIsInstance(content[7],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertEqual(len(content[1]),2)
			self.assertEqual(len(content[2]),2)
			self.assertEqual(len(content[3]),2)
			self.assertEqual(len(content[4]),2)
			self.assertEqual(len(content[5]),2)
			self.assertEqual(len(content[6]),2)
			self.assertEqual(len(content[7]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)
			self.assertIsInstance(content[1][0],int)
			self.assertIsInstance(content[1][1],str)
			self.assertIsInstance(content[2][0],int)
			self.assertIsInstance(content[2][1],str)
			self.assertIsInstance(content[3][0],int)
			self.assertIsInstance(content[3][1],str)
			self.assertIsInstance(content[4][0],int)
			self.assertIsInstance(content[4][1],str)
			self.assertIsInstance(content[5][0],int)
			self.assertIsInstance(content[5][1],str)
			self.assertIsInstance(content[6][0],int)
			self.assertIsInstance(content[6][1],str)
			self.assertIsInstance(content[7][0],int)
			self.assertIsInstance(content[7][1],str)

		for content in (

			self.prq.content(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2),
			self.prq.content(self.d_mss2_c, self.MEDIUM_SET_SPROUT_2, apparent=True),
		):
			self.assertEqual(len(content),1)
			self.assertIsInstance(content[0],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)

		for content in (

			self.prq.content(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3),
			self.prq.content(self.d_mss3_k_t, self.MEDIUM_SET_SPROUT_3, apparent=True),
		):
			self.assertEqual(len(content),1)
			self.assertIsInstance(content[0],tuple)
			self.assertEqual(len(content[0]),2)
			self.assertIsInstance(content[0][0],int)
			self.assertIsInstance(content[0][1],str)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.hedwig.mail.letter.fields	import SenderField
from	pygwarts.hedwig.mail.letter.fields	import RecipientField
from	pygwarts.hedwig.mail.letter.fields	import ccField
from	pygwarts.hedwig.mail.letter.fields	import bccField
from	pygwarts.hedwig.mail.letter.fields	import SubjectField
from	pygwarts.hedwig.mail.letter.fields	import BodyField
from	pygwarts.hedwig.mail.letter.fields	import AttachmentField
from	pygwarts.hedwig.mail.builder		import LetterBuilder
from	pygwarts.hedwig.mail.utils			import EmailValidator
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access.utils			import TextWrapper
from	pygwarts.tests.hedwig				import HedwigTestCase








class BuilderCases(HedwigTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.HEDWIG_MAIL_BUILDER_HANDLER): os.remove(cls.HEDWIG_MAIL_BUILDER_HANDLER)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.HEDWIG_MAIL_BUILDER_HANDLER)
	def test_no_mandatories_build(self):

		class CurrentBuilder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_BUILDER_HANDLER
				init_name	= "no_mandatories_build"
				init_level	= 10

		self.test_case = CurrentBuilder()
		with self.assertLogs("no_mandatories_build", 10) as case_loggy : self.test_case.build()
		self.assertEqual(len(case_loggy.output),3)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining sender field value", case_loggy.output)
		self.assertIn("DEBUG:no_mandatories_build:sender field is not a list", case_loggy.output)
		self.assertIn(

			f"CRITICAL:no_mandatories_build:{self.test_case} sender field failed",
			case_loggy.output
		)


		self.test_case("sender", [ 42 ])
		with self.assertLogs("no_mandatories_build", 10) as case_loggy : self.test_case.build()
		self.assertEqual(len(case_loggy.output),4)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining sender field value", case_loggy.output)
		self.assertIn(

			f"DEBUG:no_mandatories_build:{self.test_case['sender'][0]} is not callable",
			case_loggy.output
		)
		self.assertIn("INFO:no_mandatories_build:sender field obtained 0 items", case_loggy.output)
		self.assertIn(

			f"CRITICAL:no_mandatories_build:{self.test_case} sender field failed",
			case_loggy.output
		)


		self.test_case("sender", [ lambda : 42 ])
		with self.assertLogs("no_mandatories_build", 10) as case_loggy : self.test_case.build()
		self.assertEqual(len(case_loggy.output),4)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining sender field value", case_loggy.output)
		self.assertIn(

			f"DEBUG:no_mandatories_build:{self.test_case['sender'][0]} returned not a string",
			case_loggy.output
		)
		self.assertIn("INFO:no_mandatories_build:sender field obtained 0 items", case_loggy.output)
		self.assertIn(

			f"CRITICAL:no_mandatories_build:{self.test_case} sender field failed",
			case_loggy.output
		)


		self.test_case("sender", [ lambda : "example@email.com" ])
		with self.assertLogs("no_mandatories_build", 10) as case_loggy : self.test_case.build()
		self.assertEqual(len(case_loggy.output),5)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining sender field value", case_loggy.output)
		self.assertIn("INFO:no_mandatories_build:sender field obtained 1 item", case_loggy.output)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining recipient field value", case_loggy.output)
		self.assertIn("DEBUG:no_mandatories_build:recipient field is not a list", case_loggy.output)
		self.assertIn(

			f"CRITICAL:no_mandatories_build:{self.test_case} recipient field failed",
			case_loggy.output
		)


		self.test_case("recipient", [ lambda : 42 ])
		with self.assertLogs("no_mandatories_build", 10) as case_loggy : self.test_case.build()
		self.assertEqual(len(case_loggy.output),6)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining sender field value", case_loggy.output)
		self.assertIn("INFO:no_mandatories_build:sender field obtained 1 item", case_loggy.output)
		self.assertIn("DEBUG:no_mandatories_build:Obtaining recipient field value", case_loggy.output)
		self.assertIn(

			f"DEBUG:no_mandatories_build:{self.test_case['recipient'][0]} returned not a string",
			case_loggy.output
		)
		self.assertIn("INFO:no_mandatories_build:recipient field obtained 0 symbols", case_loggy.output)
		self.assertIn(

			f"CRITICAL:no_mandatories_build:{self.test_case} recipient field failed",
			case_loggy.output
		)




	def test_mandatories_build(self):

		class CurrentBuilder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_BUILDER_HANDLER
				init_name	= "mandatories_build"
				init_level	= 10

			class validator(EmailValidator):	pass
			class one(SenderField):				field_value = "example1@email.com"
			class two(RecipientField):			field_value = "example2@email.com"

		with self.assertLogs("mandatories_build", 10) as case_loggy:

			self.test_case = CurrentBuilder()
			self.assertIn("sender", self.test_case)
			self.assertIsInstance(self.test_case["sender"][0], SenderField)
			self.assertIn("recipient", self.test_case)
			self.assertIsInstance(self.test_case["recipient"][0], RecipientField)
			current_build = self.test_case.build()

		self.assertEqual(len(case_loggy.output),22)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:mandatories_build:Putted \"sender\", \"{[ self.test_case.one ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:mandatories_build:Putted \"recipient\", \"{[ self.test_case.two ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining sender field value", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Fetched value of {type('')}", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:mandatories_build:Fetched value of {type('')}"),2
		)
		self.assertIn(
			f"DEBUG:mandatories_build:Email address \"example1@email.com\" validated", case_loggy.output
		)
		self.assertIn(f"INFO:mandatories_build:sender field obtained 1 item", case_loggy.output)
		self.assertIn(f"INFO:mandatories_build:{self.test_case.one} obtained 1 email", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining recipient field value", case_loggy.output)
		self.assertIn(
			f"DEBUG:mandatories_build:Email address \"example2@email.com\" validated", case_loggy.output
		)
		self.assertIn(f"INFO:mandatories_build:recipient field obtained 19 symbols", case_loggy.output)
		self.assertIn(f"INFO:mandatories_build:{self.test_case.two} obtained 1 email", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining cc field value", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:cc field is not a list", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining bcc field value", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:bcc field is not a list", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining subject field value", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:subject field is not a list", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining body field value", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:body field is not a list", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:Obtaining attachment field value", case_loggy.output)
		self.assertIn(f"DEBUG:mandatories_build:attachment field is not a list", case_loggy.output)

		self.assertIsInstance(current_build, dict)
		self.assertEqual(len(current_build), 7)
		self.assertIn("sender", current_build)
		self.assertIn("recipient", current_build)
		self.assertIn("cc", current_build)
		self.assertIn("bcc", current_build)
		self.assertIn("subject", current_build)
		self.assertIn("body", current_build)
		self.assertIn("attachment", current_build)
		self.assertEqual(current_build["sender"],[ "example1@email.com;" ])
		self.assertEqual(current_build["recipient"], "example2@email.com;")
		self.assertIsNone(current_build["cc"])
		self.assertIsNone(current_build["bcc"])
		self.assertIsNone(current_build["subject"])
		self.assertIsNone(current_build["body"])
		self.assertIsNone(current_build["attachment"])








	def test_Field_outter_fail(self):
		class NotBuilder(Transmutable):

			class Field(LetterBuilder.Field):	pass
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_BUILDER_HANDLER
				init_name	= "Field_outter_fail"
				init_level	= 10


		with self.assertLogs("Field_outter_fail", 10) as case_loggy:

			self.test_case = NotBuilder()
			current = self.test_case.Field("sender")

		self.assertIsNone(current)
		self.assertIn("DEBUG:Field_outter_fail:Obtaining sender field value", case_loggy.output)
		self.assertIn(
			"DEBUG:Field_outter_fail:sender field cannot be included for building", case_loggy.output
		)








	def test_full_build(self):

		attachyone = str(self.HEDWIG_ROOT /"atachy.one")
		if	not os.path.isfile(attachyone): self.fmake(attachyone, "heil full build!")
		self.assertTrue(os.path.isfile(attachyone))

		class CurrentBuilder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_BUILDER_HANDLER
				init_name	= "full_build"
				init_level	= 10

			class validator(EmailValidator):	pass
			class one(SenderField):				field_value = "example1@email.com"
			class two(RecipientField):			field_value = "example2@email.com"
			class three(ccField):				field_value = "example3@email.com"
			class four(bccField):				field_value = "example4@email.com"
			class five(SubjectField):			field_value = "full build test"
			class six(BodyField):				field_value = "congratulations! build is nice"
			class seven(AttachmentField):		field_value = attachyone

		with self.assertLogs("full_build", 10) as case_loggy:

			self.test_case = CurrentBuilder()

			self.assertIsInstance(self.test_case["sender"], list)
			self.assertIsInstance(self.test_case["recipient"], list)
			self.assertIsInstance(self.test_case["cc"], list)
			self.assertIsInstance(self.test_case["bcc"], list)
			self.assertIsInstance(self.test_case["subject"], list)
			self.assertIsInstance(self.test_case["body"], list)
			self.assertIsInstance(self.test_case["attachment"], list)
			self.assertIsInstance(self.test_case["sender"][0], SenderField)
			self.assertIsInstance(self.test_case["recipient"][0], RecipientField)
			self.assertIsInstance(self.test_case["cc"][0], ccField)
			self.assertIsInstance(self.test_case["bcc"][0], bccField)
			self.assertIsInstance(self.test_case["subject"][0], SubjectField)
			self.assertIsInstance(self.test_case["body"][0], BodyField)
			self.assertIsInstance(self.test_case["attachment"][0], AttachmentField)

			current_build = self.test_case.build()

		self.assertEqual(len(case_loggy.output),45)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:full_build:Putted \"sender\", \"{[ self.test_case.one ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:full_build:Putted \"recipient\", \"{[ self.test_case.two ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:full_build:Putted \"cc\", \"{[ self.test_case.three ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:full_build:Putted \"bcc\", \"{[ self.test_case.four ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:full_build:Putted \"subject\", \"{[ self.test_case.five ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:full_build:Putted \"body\", \"{[ self.test_case.six ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:full_build:Putted \"attachment\", \"{[ self.test_case.seven ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:full_build:Obtaining sender field value", case_loggy.output)
		self.assertIn(f"DEBUG:full_build:Fetched value of {type('')}", case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:full_build:Fetched value of {type('')}"),7)
		self.assertIn(
			"DEBUG:full_build:Email address \"example1@email.com\" validated", case_loggy.output
		)
		self.assertIn("INFO:full_build:sender field obtained 1 item", case_loggy.output)
		self.assertIn(f"INFO:full_build:{self.test_case.one} obtained 1 email", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtaining recipient field value", case_loggy.output)
		self.assertIn(
			"DEBUG:full_build:Email address \"example2@email.com\" validated", case_loggy.output
		)
		self.assertIn("INFO:full_build:recipient field obtained 19 symbols", case_loggy.output)
		self.assertIn(f"INFO:full_build:{self.test_case.two} obtained 1 email", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtaining cc field value", case_loggy.output)
		self.assertIn(
			"DEBUG:full_build:Email address \"example3@email.com\" validated", case_loggy.output
		)
		self.assertIn("INFO:full_build:cc field obtained 19 symbols", case_loggy.output)
		self.assertIn(f"INFO:full_build:{self.test_case.three} obtained 1 email", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtaining bcc field value", case_loggy.output)
		self.assertIn(
			"DEBUG:full_build:Email address \"example4@email.com\" validated", case_loggy.output
		)
		self.assertIn("INFO:full_build:bcc field obtained 19 symbols", case_loggy.output)
		self.assertIn(f"INFO:full_build:{self.test_case.four} obtained 1 email", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtaining subject field value", case_loggy.output)
		self.assertIn("DEBUG:full_build:Proper modifiers not found", case_loggy.output)
		self.assertEqual(case_loggy.output.count("DEBUG:full_build:Proper modifiers not found"), 2)
		self.assertIn("DEBUG:full_build:Fetched 15 symbols", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtained 15 symbols", case_loggy.output)
		self.assertIn("INFO:full_build:subject field obtained 15 symbols", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtaining body field value", case_loggy.output)
		self.assertIn("DEBUG:full_build:Fetched 30 symbols", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtained 30 symbols", case_loggy.output)
		self.assertIn("INFO:full_build:body field obtained 30 symbols", case_loggy.output)
		self.assertIn("DEBUG:full_build:Obtaining attachment field value", case_loggy.output)
		self.assertIn(f"DEBUG:full_build:Fetched {len(attachyone)} symbols", case_loggy.output)
		self.assertIn("DEBUG:full_build:Modifiers not used", case_loggy.output)
		self.assertIn(f"DEBUG:full_build:Obtained file path \"{attachyone}\"", case_loggy.output)
		self.assertIn(
			f"INFO:full_build:attachment field obtained 1 item", case_loggy.output
		)
		self.assertIsInstance(current_build, dict)
		self.assertEqual(len(current_build), 7)
		self.assertIn("sender", current_build)
		self.assertIn("recipient", current_build)
		self.assertIn("cc", current_build)
		self.assertIn("bcc", current_build)
		self.assertIn("subject", current_build)
		self.assertIn("body", current_build)
		self.assertIn("attachment", current_build)
		self.assertEqual(current_build["sender"],[ "example1@email.com;" ])
		self.assertEqual(current_build["recipient"], "example2@email.com;")
		self.assertEqual(current_build["cc"], "example3@email.com;")
		self.assertEqual(current_build["bcc"], "example4@email.com;")
		self.assertEqual(current_build["subject"], "full build test")
		self.assertEqual(current_build["body"], "congratulations! build is nice")
		self.assertEqual(current_build["attachment"],[ attachyone ])

		if	os.path.isfile(attachyone): os.remove(attachyone)
		self.assertFalse(os.path.isfile(attachyone))








	def test_maximum_build(self):

		attachytwo = str(self.HEDWIG_ROOT /"atachy.two")
		if	not os.path.isfile(attachytwo): self.fmake(attachytwo, "heil maximum build!")
		self.assertTrue(os.path.isfile(attachytwo))

		class CurrentBuilder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_BUILDER_HANDLER
				init_name	= "maximum_build"
				init_level	= 10

			class validator(EmailValidator):	pass
			class one(SenderField):				field_value = "example1@email.com", "exomple1@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			class three(ccField):				field_value = "example4@email.com; example5@email.com"
			class four(bccField):				field_value = "example6@email.com; example7@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

			class seven(AttachmentField):

				field_value	= "attachytwo"
				modifiers	= lambda *S : attachytwo

		with self.assertLogs("maximum_build", 10) as case_loggy:

			self.test_case = CurrentBuilder()

			self.assertIsInstance(self.test_case["sender"], list)
			self.assertIsInstance(self.test_case["recipient"], list)
			self.assertIsInstance(self.test_case["cc"], list)
			self.assertIsInstance(self.test_case["bcc"], list)
			self.assertIsInstance(self.test_case["subject"], list)
			self.assertIsInstance(self.test_case["body"], list)
			self.assertIsInstance(self.test_case["attachment"], list)
			self.assertIsInstance(self.test_case["sender"][0], SenderField)
			self.assertIsInstance(self.test_case["recipient"][0], RecipientField)
			self.assertIsInstance(self.test_case["cc"][0], ccField)
			self.assertIsInstance(self.test_case["bcc"][0], bccField)
			self.assertIsInstance(self.test_case["subject"][0], SubjectField)
			self.assertIsInstance(self.test_case["body"][0], BodyField)
			self.assertIsInstance(self.test_case["attachment"][0], AttachmentField)

			current_build = self.test_case.build()

		self.assertEqual(len(case_loggy.output),48)
		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"sender\", \"{[ self.test_case.one ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"recipient\", \"{[ self.test_case.two ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"cc\", \"{[ self.test_case.three ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"bcc\", \"{[ self.test_case.four ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"subject\", \"{[ self.test_case.five ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"body\", \"{[ self.test_case.six ]}\" pair",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:maximum_build:Putted \"attachment\", \"{[ self.test_case.seven ]}\" pair",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:maximum_build:Obtaining sender field value", case_loggy.output)
		self.assertIn(f"DEBUG:maximum_build:Fetched value of {type(tuple())}", case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:maximum_build:Fetched value of {type(tuple())}"),2)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example1@email.com\" validated", case_loggy.output
		)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"exomple1@email.com\" validated", case_loggy.output
		)
		self.assertIn(f"INFO:maximum_build:{self.test_case.one} obtained 2 emails", case_loggy.output)
		self.assertIn("INFO:maximum_build:sender field obtained 1 item", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtaining recipient field value", case_loggy.output)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example2@email.com\" validated", case_loggy.output
		)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example3@email.com\" validated", case_loggy.output
		)
		self.assertIn(f"INFO:maximum_build:{self.test_case.two} obtained 2 emails", case_loggy.output)
		self.assertIn("INFO:maximum_build:recipient field obtained 38 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtaining cc field value", case_loggy.output)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example4@email.com\" validated", case_loggy.output
		)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example5@email.com\" validated", case_loggy.output
		)
		self.assertIn(f"INFO:maximum_build:{self.test_case.three} obtained 2 emails", case_loggy.output)
		self.assertIn("INFO:maximum_build:cc field obtained 38 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtaining bcc field value", case_loggy.output)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example6@email.com\" validated", case_loggy.output
		)
		self.assertIn(
			"DEBUG:maximum_build:Email address \"example7@email.com\" validated", case_loggy.output
		)
		self.assertIn(f"INFO:maximum_build:{self.test_case.four} obtained 2 emails", case_loggy.output)
		self.assertIn("INFO:maximum_build:bcc field obtained 38 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtaining subject field value", case_loggy.output)
		self.assertIn(f"DEBUG:maximum_build:Fetched value of {type('')}", case_loggy.output)
		self.assertEqual(case_loggy.output.count(f"DEBUG:maximum_build:Fetched value of {type('')}"),5)
		self.assertIn("DEBUG:maximum_build:Fetched 18 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtained 18 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Wrapped text now 27 symbols", case_loggy.output)
		self.assertIn("INFO:maximum_build:subject field obtained 27 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtaining body field value", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Fetched 35 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtained 30 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Wrapped text now 40 symbols", case_loggy.output)
		self.assertIn("INFO:maximum_build:body field obtained 40 symbols", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Obtaining attachment field value", case_loggy.output)
		self.assertIn("DEBUG:maximum_build:Fetched 10 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:maximum_build:Obtained file path \"{attachytwo}\"", case_loggy.output)
		self.assertIn(
			f"INFO:maximum_build:attachment field obtained 1 item", case_loggy.output
		)
		self.assertIsInstance(current_build, dict)
		self.assertEqual(len(current_build), 7)
		self.assertIn("sender", current_build)
		self.assertIn("recipient", current_build)
		self.assertIn("cc", current_build)
		self.assertIn("bcc", current_build)
		self.assertIn("subject", current_build)
		self.assertIn("body", current_build)
		self.assertIn("attachment", current_build)
		self.assertEqual(current_build["sender"],[ "example1@email.com;exomple1@email.com;" ])
		self.assertEqual(current_build["recipient"], "example2@email.com;example3@email.com;")
		self.assertEqual(current_build["cc"], "example4@email.com;example5@email.com;")
		self.assertEqual(current_build["bcc"], "example6@email.com;example7@email.com;")
		self.assertEqual(current_build["subject"], "Subject: maximum build test")
		self.assertEqual(current_build["body"], "My congratulations! Build is nice, baby!")
		self.assertEqual(current_build["attachment"],[ attachytwo ])

		if	os.path.isfile(attachytwo): os.remove(attachytwo)
		self.assertFalse(os.path.isfile(attachytwo))








	def test_multiple_fields_build(self):

		f1 = str(self.HEDWIG_ROOT /"f1.two")
		f2 = str(self.HEDWIG_ROOT /"f2.two")
		f3 = str(self.HEDWIG_ROOT /"f3.two")
		f4 = str(self.HEDWIG_ROOT /"f4.two")
		f5 = str(self.HEDWIG_ROOT /"f5.two")

		if	not os.path.isfile(f1): self.fmake(f1, "OOH EEH")
		if	not os.path.isfile(f2): self.fmake(f2, "OOH AH AH")
		if	not os.path.isfile(f3): self.fmake(f3, "TING TANG")
		if	not os.path.isfile(f4): self.fmake(f4, "WALLA WALLA")
		if	not os.path.isfile(f5): self.fmake(f5, "BING BANG")

		self.assertTrue(os.path.isfile(f1))
		self.assertTrue(os.path.isfile(f2))
		self.assertTrue(os.path.isfile(f3))
		self.assertTrue(os.path.isfile(f4))
		self.assertTrue(os.path.isfile(f5))

		class CurrentBuilder(LetterBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_BUILDER_HANDLER
				init_name	= "multiple_fields_build"
				init_level	= 10

			class validator(EmailValidator):	pass
			class from1(SenderField):			field_value = "example1@email.com"
			class from2(SenderField):			field_value = "example2@email.com"
			class to1(RecipientField):			field_value = "example3@email.com;example4@email.com;"
			class to2(RecipientField):			field_value = "example5@email.com", "example6@email.com"
			class with1(ccField):				field_value = [ "example7@email.com", "example8@email.com" ]
			class with2(ccField):				field_value = { "example9@email.com" }
			class also1(bccField):				field_value = { "e1@m.com": 1, "e2@m.com": 1 }
			class also2(bccField):				field_value = "example10@email.com"
			class also3(bccField):				field_value = "example11@email.com"
			@TextWrapper("QTC: ")
			@TextWrapper("Subject: ")
			class subj1(SubjectField):

				field_value	= "{level} build"
				modifiers	= { "level": "maximum" }

			class subj2(SubjectField):	field_value = " test"
			class subj3(SubjectField):	field_value = "!"
			@TextWrapper(header="My ", footer=", baby!")
			class b1(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

			class b2(BodyField):		field_value = "\n\nauthor: me"
			class b3(BodyField):		field_value = " (no doubt)"
			class A1(AttachmentField):	field_value	= f1
			class A2(AttachmentField):	field_value	= f2
			class A3(AttachmentField):	field_value	= f3
			class A4(AttachmentField):	field_value	= f4
			class A5(AttachmentField):	field_value	= f5


		with self.assertLogs("multiple_fields_build", 10) as case_loggy:

			self.test_case = CurrentBuilder()

			self.assertIsInstance(self.test_case["sender"], list)
			self.assertIsInstance(self.test_case["recipient"], list)
			self.assertIsInstance(self.test_case["cc"], list)
			self.assertIsInstance(self.test_case["bcc"], list)
			self.assertIsInstance(self.test_case["subject"], list)
			self.assertIsInstance(self.test_case["body"], list)
			self.assertIsInstance(self.test_case["attachment"], list)
			self.assertEqual(len(self.test_case["sender"]),2)
			self.assertIsInstance(self.test_case["sender"][0], SenderField)
			self.assertIsInstance(self.test_case["sender"][1], SenderField)
			self.assertEqual(len(self.test_case["recipient"]),2)
			self.assertIsInstance(self.test_case["recipient"][0], RecipientField)
			self.assertIsInstance(self.test_case["recipient"][1], RecipientField)
			self.assertEqual(len(self.test_case["cc"]),2)
			self.assertIsInstance(self.test_case["cc"][0], ccField)
			self.assertIsInstance(self.test_case["cc"][1], ccField)
			self.assertEqual(len(self.test_case["bcc"]),3)
			self.assertIsInstance(self.test_case["bcc"][0], bccField)
			self.assertIsInstance(self.test_case["bcc"][1], bccField)
			self.assertIsInstance(self.test_case["bcc"][2], bccField)
			self.assertEqual(len(self.test_case["subject"]),3)
			self.assertIsInstance(self.test_case["subject"][0], SubjectField)
			self.assertIsInstance(self.test_case["subject"][1], SubjectField)
			self.assertIsInstance(self.test_case["subject"][2], SubjectField)
			self.assertEqual(len(self.test_case["body"]),3)
			self.assertIsInstance(self.test_case["body"][0], BodyField)
			self.assertIsInstance(self.test_case["body"][1], BodyField)
			self.assertIsInstance(self.test_case["body"][2], BodyField)
			self.assertEqual(len(self.test_case["attachment"]),5)
			self.assertIsInstance(self.test_case["attachment"][0], AttachmentField)
			self.assertIsInstance(self.test_case["attachment"][1], AttachmentField)
			self.assertIsInstance(self.test_case["attachment"][2], AttachmentField)
			self.assertIsInstance(self.test_case["attachment"][3], AttachmentField)
			self.assertIsInstance(self.test_case["attachment"][4], AttachmentField)

			current_build = self.test_case.build()

		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIsInstance(current_build, dict)
		self.assertEqual(len(current_build), 7)
		self.assertIn("sender", current_build)
		self.assertIn("recipient", current_build)
		self.assertIn("cc", current_build)
		self.assertIn("bcc", current_build)
		self.assertIn("subject", current_build)
		self.assertIn("body", current_build)
		self.assertIn("attachment", current_build)
		self.assertEqual(current_build["sender"],[ "example1@email.com;", "example2@email.com;" ])
		self.assertEqual(

			current_build["recipient"],
			"example3@email.com;example4@email.com;example5@email.com;example6@email.com;"
		)
		self.assertEqual(

			current_build["cc"],
			"example7@email.com;example8@email.com;example9@email.com;"
		)
		self.assertEqual(

			current_build["bcc"],
			"e1@m.com;e2@m.com;example10@email.com;example11@email.com;"
		)
		self.assertEqual(current_build["subject"], "QTC: Subject: maximum build test!")
		self.assertEqual(

			current_build["body"],
			"My congratulations! Build is nice, baby!\n\nauthor: me (no doubt)"
		)
		self.assertEqual(current_build["attachment"],[ f1, f2, f3, f4, f5 ])

		if	os.path.isfile(f1): os.remove(f1)
		if	os.path.isfile(f2): os.remove(f2)
		if	os.path.isfile(f3): os.remove(f3)
		if	os.path.isfile(f4): os.remove(f4)
		if	os.path.isfile(f5): os.remove(f5)

		self.assertFalse(os.path.isfile(f1))
		self.assertFalse(os.path.isfile(f2))
		self.assertFalse(os.path.isfile(f3))
		self.assertFalse(os.path.isfile(f4))
		self.assertFalse(os.path.isfile(f5))








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing							import List
from typing							import Dict
from smtplib						import SMTP_SSL
from pygwarts.magical.spells		import patronus
from pygwarts.hedwig.mail.builder	import LetterBuilder








class SMTPBuilder(LetterBuilder):

	"""
		LetterBuilder implementation of standard SMTP library. Initiates SSL connection and server
		authorization, using credentials that must be provided as a dictionary for "build" method,
		preliminary verifying them with "is_connectable" helper method. Fetches letter fields by
		LetterBuilder.build and sends email only for a very first SenderField entity encountered
		immediately. Ignores all attachments. All cc and bcc fields extends recipient field. Handles
		every possible Exception that might be raised during connection, authorization and sending.
	"""

	def is_connectable(self, probe :Dict[str,str | int]) -> bool :

		"""
			Helper method that verifies connection "probe" dictionary. Checks if such dictionary has
			such necessary key-value pairs as:
				"endpoint":	email server endpoint address string,
				"port":		email server address port as integer,
				"password":	password for authorization on the server with very first SenderField address
							as login.
		"""

		valid = True

		if	isinstance(probe, dict):
			if	not isinstance(probe.get("endpoint"), str):

				self.loggy.info(f"Invalid {self} endpoint type")
				valid = False


			if	not isinstance(probe.get("port"), int):

				self.loggy.info(f"Invalid {self} port type")
				valid = False


			if	not isinstance(probe.get("password"), str):

				self.loggy.info(f"Invalid {self} password type")
				valid = False


			if		valid : self.loggy.debug("Credentials probe validated")
			return	valid




	def build(self, credentials :Dict[str,str | int]) -> Dict[str,str | List[str]] :

		"""
			Core method for LetterBuilder implementation of standard SMTP library. Accepts "credentials"
			dictionary with email server connection information, which must include such key-value pairs as:
				"endpoint":	email server endpoint address string,
				"port":		email server address port as integer,
				"password":	password for authorization on the server with very first SenderField address
							as login.
			Returns fetched by LetterBuilder.build letter dictionary in case letter was sent to every
			recipient, which means SMTP connection response for a letter sending was not a dictionary with
			rejected recipients entities.
		"""

		if	self.is_connectable(credentials):
			if	self.is_buildable(current_build := super().build()):

				current_sender	= current_build["sender"][0].rstrip(';')
				current_recip	= current_build["recipient"]
				current_subject	= current_build["subject"] or str()
				current_body	= current_build["body"] or str()
				current_cc		= current_build["cc"]
				current_bcc		= current_build["bcc"]

				if	current_cc is not None	: current_recip += current_cc
				if	current_bcc is not None	: current_recip += current_bcc

				try:

					self.loggy.info(f"{self} commence {current_sender} letter build")
					self.current_smtp_connection = SMTP_SSL(

						credentials["endpoint"],
						credentials["port"],
						timeout=30,
					)
					self.loggy.debug(f"SMTP connection established")


					login = self.current_smtp_connection.login(current_sender, credentials["password"])
					self.loggy.debug(f"Login status: {login}")


					response = self.current_smtp_connection.sendmail(

						current_sender,
						[ address.strip() for address in current_recip.split(";") if address ],
						f"From: {current_sender}\r\n"
						f"To: {current_recip}\r\n"
						f"Subject: {current_subject}\r\n\r\n"
						f"{current_body}"
					)


				except	Exception as E : self.loggy.info(f"{self} failed due to {patronus(E)}")
				else:
					if	isinstance(response, dict):
						if	len(response):
							for recipient,details in response.items():

								# Method "smtp.sendmail" will return normally if the mail is accepted for
								# at least one recipient. Otherwise it will raise an exception. If this
								# method does not raise an exception, it returns a dictionary, with one
								# entry for each recipient that was refused. Each entry contains a tuple of
								# the SMTP error code and the accompanying error message sent by the server.
								self.loggy.info(f"{self} didn't sent to {recipient} {details}")
						else:

							self.loggy.info(f"{self} all recipients accepted")
							return	current_build
					else:	self.loggy.debug(f"Unexpected response \"{response}\"")
				finally:
						if	hasattr(self, "current_smtp_connection") : self.current_smtp_connection.quit()










from __future__								import annotations
from typing									import Any
from typing									import List
from typing									import Hashable
from pygwarts.magical.philosophers_stone	import Transmutable








class Chest(Transmutable):

	"""
				         ..~>%######%%%??!!>;:=~-,....                                                    
		   .._=>?#&?<::::::::::::::;><<!?%#%%??!!>;:++=~---___----~~~~--__,,....... ....          
		.-:%@B$%??!!!!!!???????????!!!<<>>;;:::::;>><!!???%%%%???!!!!!??%%%%%###%%%%%????!!!!<<>+.
		@#!<<>;;::::::::::::::::::::;;;>>><<!!!!!!?????!!<<<<<<>>>;::::::::::::::::::><?%%#&&##&@.
		@?::::::;;;;::::::::::::::::::::::::::::::::::;;>>>><<<!!!??????????!<>!?%##&#%????????#@.
		?&:#?<;>@!!!??#:::;!?<:::::<!!>:::::::::::::::::::::::::::::::::::::;#&%????????????&&?&#.
		>@:%!#<:@;%?#!$:::&>;?#:::>#<:#>::::::::::<%?%!::!????;::::;>>>;:::::!&????#&&#???#%$&?$>.
		+B::::::@;<?!;$:::><::!#%!<<!?%;::::::::::@>??>:<#:::>@>::>&<<<<!&;>?;$?%&?%@??????##?&B?.
		=B#%%%%%B%%???@!!<<>>;;::>>>;:::::::::::::>%?!<?&;::!!%;:::&>%?&;&;??;#%?&&##?????%#$@#%B>
		~B;:::::!#;;;$!>><<!!???????????!!<>;::::::::;>;::::::::::;#;?!!;&::::?&??????%#&&@%;@??&#
		~B;:::::!%;;;@::::::::::::::::::;><!!?????????!!<<>>;;::::<%;;;;;&::::#&?%#&$$#%??#%;@%##$
		~B::::::<%;;;$::::::::::::::::::::::::::::::::;>><<!!??????$$%%%%@%%##$@&&$>%#???%&#;!?@#@
		=B::::::<#;;>&::::::::::::::::>?#%%%%????>:::::::::::::::::<$;;;>&::::<$??B;##?%%%#?;<&?%$
		=B:::##??<;;>#???%%>:::::::::##<;;;;;;;;>%&>:::::::::::::::;@;;;<%::::>$??@?>?!?B%?%&&??%&
		+@:::;%#>;;;;;;;!$!:::::::::>$;;;;;;;;;;<;<@:::::::::::::::;B;;;%!::::>@?##>;;>$%???????##
		+B:::::;?%>;;;<&?::::::::::;$#<;;;;;;;;;;;##:::::::::::::%%?>;;;!%??<:;@?%&?><$%????????#%
		+B:::::::;?#!%?;:::::::::::;@!;;;;;;;;;;;;!%%;:::::::::::<&?;;;;;;?&<::B???%#&??????%#%?&?
		=B::::::::::<>::::::::::::%$!;;;;;;;;;;;;;%#?;:::::::::::::!&!;;!&!::::@????????????&#$?#?
		-B;::::::::::::::::::::::::<%%!;;;;;;;;;;;?#!::::::::::::::::!##<::::::$????%????????%??#?
		.B>::::::::$#$$::::::::::::::%#;;;;;;;;;;!%%$<:::::::::::::::::::::::::&%???$$#????%$&&?#%
		.$<::::::::>??>:::::::::::::>@%%%#!;?%!;!#::::::::::::::::::::<!>::::::#%???%%???????%??##
		.#!:::::::::::::::::::::::::::::::##<:>?%::::::::::::::::::::<$%@>:::::%#???###?????%???##
		.?%:::::::;<>:::::::::::::::::::::::::::::::::::::::::::::::::><>::::::?&???###????%&!#$#$
		.<#:::::::<$$&:::::::::::>!!<>>>><<<<<<<<<<<<<<<<>>>:::::::::::::::::::!&??????????#;!@?#B
		.;$::::::::;;::::::::::::;<<!!<<<<<<>>>>>>>>>>><<<<<>::::::::::::::::::>$?????????$??;%$B?
		 :@::::::::<?%>::::::::;%%%?!!<<<>>>>>>>>>>>>>>;;;:::::::::::::%%!;::::;@??%&%??????&<<B>.
		 =B;::::;%#!;>%#>:::::::::;>><<!!!!!!!!!!!!!!!????%#!::::::::::%?#<:::::@??#!><!?&%?#$&-. 
		 _B<:::;@%>;;;;;%%;:::::::::::::::::::::::::::::::::::::::::::::>>::::::@??@!%>>&#?%@?..  
		 .@!::::;>##;;>%?%@<::::::::>#%%>::::::::::::::::::::::::::::;%#!?#<::::$%?%%@>>@?#@:.    
		 .#%::::::&!;;;@;;>>:>?!!:::;?><@::::::::;&%#<:::>??%;:::::>##>;;;;!#<::#%???$!;#B&_.     
		 .<&;::::;B;;;;@:::::#;:<%?!<!%#<:::::::::&!<;:<#?:>&?::::>B&%!;;;>!?@!:%#???&&>$!.       
		  ,;?%###$@??!!$!>>;;::::::;>;:::::::::::::!%%%!;::;;::::::::&%;;;?#;:::?&????B@;.        
		     ...,_-~=++:;;><<!???%%%###%%??!!<<>>>;;:::::::::::::::::@>;;;%!::::!&??%$&~.         
		                        ........,__-~~==+::;;>>>><<<<!!!!???#@%###@&##%%#&%?>~..          
		Mutable chain implementation of container with Transmutables. It is a special list, as all logic
		lies behind maintaining "_inside_" list, to be a part of mutable chain.
		Any Transmutables can be puted into Chest by calling the Chest, providing them as argument.
		Providing no arguments during Chest call will simply return a copy of "_inside_" list
		of Transmutables (calling with arguments to put Transmutables in Chest will also return such copy).
		Chest can be iterated, compared to be equal to another Chest or a list, compared to be
		greater (or equal)/lesser (or equal) than another Chest or list (comparison will take place
		against lengths), Transmutable items can be assigned/obtained/deleted. The built-in function
		"next" will take place instantly, so Chest is iterator by default. There is no "full exhaustion"
		of such iterator, so after StopIteration raised when all "_inside_" elements passed, the next
		call of "next" (after call that raised StopIteration), the iteration starts over again.
	"""

	def __len__(self) -> int : return len(self._inside_)
	def __hash__(self) -> int : return id(self)
	def __contains__(self, item :Transmutable) -> bool : return item in self._inside_
	def __getitem__	(self, index :int) -> Transmutable | None :

		"""
			Trying to index into "_inside_" list to get corresponding Transmutable.
			If index is out of "_inside_" range or "_inside_" length is 0, there'll be no Exception
			and simply None returned.
		"""

		if	len(self._inside_) and (0 <= index < len(self._inside_) or 0 >= index >= -len(self._inside_)):
			return self._inside_[index]




	def __setitem__(self, index :int, item :Transmutable):

		"""
			Assigning "_inside_" index to a Transmutable item. If item is not Transmutable or index
			is out of "_inside_" range, no modification to "_inside_" will be made and corresponding
			warning message will be logged with no exception raising.
		"""

		if	isinstance(item, Transmutable):
			if	index < len(self._inside_):

				self._inside_[index] = item
				self.loggy.debug(f"{item} packed at index {index}")


			else:	self.loggy.warning(f"Index {index} is out of {self} Chest range, {item} not putted")
		else:		self.loggy.warning(f"Can't put \"{type(item)}\" in {self} Chest")




	def __delitem__(self, index :int):

		"""
			Removing Transmutable from "_inside_" list. If index out of "_inside_" range, corresponding
			warning message will be logged with no "_inside_" modification and no exception raising.
		"""

		if	len(self._inside_) and (0 <= index < len(self._inside_) or 0 >= index >= -len(self._inside_)):

			item = self._inside_[index]
			del self._inside_[index]

			self.loggy.info(f"Removed {item} from {self} Chest")


		else:
			self.loggy.warning(f"Index {index} out of {self} Chest range, no removing made")




	def __eq__(self, other :Chest | List[Transmutable]) -> bool :

		"""
			Comparison of equality of current Chest "_inside_" list with another
			Chest object's "_inside_" list or with any other list object.
		"""

		match other:

			case Chest()	: return	self._inside_ == other._inside_
			case list()		: return	self._inside_ == other
			case _			: return	False




	def __gt__(self, other :Chest | List[Transmutable]) -> bool | TypeError :

		"""
			Comparison of greatness of current Chest "_inside_" list length with another Chest object's
			"_inside_" list or with any other list object's length.
		"""

		match other:

			case Chest()	: return	len(self._inside_) > len(other._inside_)
			case list()		: return	len(self._inside_) > len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} Chest")




	def __ge__(self, other :Chest | List[Transmutable]) -> bool | TypeError :

		"""
			Comparison of greatness or equality of current Chest "_inside_" list length with another
			Chest object's "_inside_" list or with any other list object's length.
		"""

		match other:

			case Chest()	: return	len(self._inside_) >= len(other._inside_)
			case list()		: return	len(self._inside_) >= len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} Chest")




	def __lt__(self, other :Chest | List[Transmutable]) -> bool | TypeError :

		"""
			Comparison of lesserness of current Chest "_inside_" list length with another Chest object's
			"_inside_" list or with any other list object's length.
		"""

		match other:

			case Chest()	: return	len(self._inside_) < len(other._inside_)
			case list()		: return	len(self._inside_) < len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} Chest")




	def __le__(self, other :Chest | List[Transmutable]) -> bool | TypeError :

		"""
			Comparison of lesserness or equality of current Chest "_inside_" list length with another
			Chest object's "_inside_" list or with any other list object's length.
		"""

		match other:

			case Chest()	: return	len(self._inside_) <= len(other._inside_)
			case list()		: return	len(self._inside_) <= len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} Chest")




	def __iter__(self):

		"""
			Iterator implementation. Logs debug message before every item yield. If there is no items
			in "_inside_", corresponding debug message will be logged.
		"""

		if	len(self._inside_):
			for item in self._inside_:

				self.loggy.debug(f"Yielding \"{item}\"")
				yield item
		else:	self.loggy.debug(f"Chest is empty")




	def __next__(self):

		"""
			Chest the iterator by default, so calling "next" will iterate through "_inside_".
			Apart to generator like iteration, Chest iteration will be reset every time StopIteration
			is raised after "_inside_" exhausted, so it might be iterated over and over again. This is
			possible by the "_current" field which tracks the "_inside_" length and resets every time
			the "_inside_" exhausted. To make "next" calls work classic way, it is possible to call it
			over "iter" call on current Chest.
		"""

		if	self._current < len(self._inside_):

			_next_ = self._inside_[self._current]
			self._current += 1


			return _next_


		self._current = 0
		raise StopIteration




	def __call__(self, *load) -> List[Transmutable,] | None:

		"""
			When single Transmutable provided in load, it is "_inside_" adding process.
			More than one argument in load will cause to check every argument to be Transmutable and put
			it to the Chest, discarding non Transmutables.
			With no arguments it is simply "_inside_" copy returning, cause __str__ will never return
			"_inside_", so such call with no arguments is the way to obtain "_inside_" list copy.
			Any other situation will cause no Chest modification, corresponding warning message
			and None as the return value, that should inform about problems.
		"""

		match load:
			case ( Transmutable(), ):

				self._inside_.append(*load)
				self.loggy.debug(f"Putted \"{load[0]}\"")

			case ( *multiple_load, ) if 1 <len(multiple_load):

				for item in multiple_load:
					if	isinstance(item, Transmutable):


						self._inside_.append(item)
						self.loggy.debug(f"Putted \"{item}\"")
					else:
						self.loggy.debug(f"Discarding \"{item}\" load")
			case ()	:	pass
			case _	:

				self.loggy.warning(f"Can't put \"{load[0]}\" in {self} Chest")
				return


		# Additional wrapping of "_inside_" to a list will cause no modification of current Chest
		# "_inside_" directly through this __call__ return value, as any modifications are supposed
		# to be maintained only with Chest functionality.
		return list(self._inside_)




	def __init__(self, *args, **kwargs):

		# Initiating Chest core fields first allows putting during initiation.
		self.unload()
		super().__init__(*args, **kwargs)




	def unload(self):

		"""
			Wipes "_inside_" items completely by reinitiating with "_current".
			Also used as initiation by itself.
		"""

		self._inside_ = list()
		self._current = 0





	def index(self, item :Transmutable) -> int | None :

		"""
			Find item in "_inside_" list and return it's first occurrence index. If item not present in
			"_inside_", or if item is not Transmutable, None will be returned and corresponding debug
			messages will be logged.
		"""

		if	isinstance(item, Transmutable):
			for i,inner in enumerate(self._inside_):
				if	inner == item:

					self.loggy.debug(f"Found {item} at index {i}")
					return i
			else:	self.loggy.debug(f"Item {item} not found")
		else:		self.loggy.debug(f"Item {item} type \"{type(item)}\" cannot be searched")








class KeyChest(Transmutable):

	"""
				         ..~>%######%%%??!!>;:=~-,....                                                    
		   .._=>?#&?<::::::::::::::;><<!?%#%%??!!>;:++=~---___----~~~~--__,,....... ....          
		.-:%@B$%??!!!!!!???????????!!!<<>>;;:::::;>><!!???%%%%???!!!!!??%%%%%###%%%%%????!!!!<<>+.
		@#!<<>;;::::::::::::::::::::;;;>>><<!!!!!!?????!!<<<<<<>>>;::::::::::::::::::><?%%#&&##&@.
		@?::::::;;;;::::::::::::::::::::::::::::::::::;;>>>><<<!!!??????????!<>!?%##&#%????????#@.
		?&:#?<;>@!!!??#:::;!?<:::::<!!>:::::::::::::::::::::::::::::::::::::;#&%????????????&&?&#.
		>@:%!#<:@;%?#!$:::&>;?#:::>#<:#>::::::::::<%?%!::!????;::::;>>>;:::::!&????#&&#???#%$&?$>.
		+B::::::@;<?!;$:::><::!#%!<<!?%;::::::::::@>??>:<#:::>@>::>&<<<<!&;>?;$?%&?%@??????##?&B?.
		=B#%%%%%B%%???@!!<<>>;;::>>>;:::::::::::::>%?!<?&;::!!%;:::&>%?&;&;??;#%?&&##?????%#$@#%B>
		~B;:::::!#;;;$!>><<!!???????????!!<>;::::::::;>;::::::::::;#;?!!;&::::?&??????%#&&@%;@??&#
		~B;:::::!%;;;@::::::::::::::::::;><!!?????????!!<<>>;;::::<%;;;;;&::::#&?%#&$$#%??#%;@%##$
		~B::::::<%;;;$::::::::::::::::::::::::::::::::;>><<!!??????$$%%%%@%%##$@&&$>%#???%&#;!?@#@
		=B::::::<#;;>&::::::::::::::::>?#%%%%????>:::::::::::::::::<$;;;>&::::<$??B;##?%%%#?;<&?%$
		=B:::##??<;;>#???%%>:::::::::##<;<!???!;>%&>:::::::::::::::;@;;;<%::::>$??@?>?!?B%?%&&??%&
		+@:::;%#>;;;;;;;!$!:::::::::>$;;@@$$&&$B<;<@:::::::::::::::;B;;;%!::::>@?##>;;>$%???????##
		+B:::::;?%>;;;<&?::::::::::;$#<;%@$&&@B@>;##:::::::::::::%%?>;;;!%??<:;@?%&?><$%????????#%
		+B:::::::;?#!%?;:::::::::::;@!;;;>@&&@$;;;!%%;:::::::::::<&?;;;;;;?&<::B???%#&??????%#%?&?
		=B::::::::::<>::::::::::::%$!;;;;;B&&&B>;;%#?;:::::::::::::!&!;;!&!::::@????????????&#$?#?
		-B;::::::::::::::::::::::::<%%!;;;$@$$$>;;?#!::::::::::::::::!##<::::::$????%????????%??#?
		.B>::::::::$#$$::::::::::::::%#;;;;;;;;;;!%%$<:::::::::::::::::::::::::&%???$$#????%$&&?#%
		.$<::::::::>??>:::::::::::::>@%%%#!;?%!;!#::::::::::::::::::::<!>::::::#%???%%???????%??##
		.#!:::::::::::::::::::::::::::::::##<:>?%::::::::::::::::::::<$%@>:::::%#???###?????%???##
		.?%:::::::;<>:::::::::::::::::::::::::::::::::::::::::::::::::><>::::::?&???###????%&!#$#$
		.<#:::::::<$$&:::::::::::>!!<>>>><<<<<<<<<<<<<<<<>>>:::::::::::::::::::!&??????????#;!@?#B
		.;$::::::::;;::::::::::::;<<!!<<<<<<>>>>>>>>>>><<<<<>::::::::::::::::::>$?????????$??;%$B?
		 :@::::::::<?%>::::::::;%%%?!!<<<>>>>>>>>>>>>>>;;;:::::::::::::%%!;::::;@??%&%??????&<<B>.
		 =B;::::;%#!;>%#>:::::::::;>><<!!!!!!!!!!!!!!!????%#!::::::::::%?#<:::::@??#!><!?&%?#$&-. 
		 _B<:::;@%>;;;;;%%;:::::::::::::::::::::::::::::::::::::::::::::>>::::::@??@!%>>&#?%@?..  
		 .@!::::;>##;;>%?%@<::::::::>#%%>::::::::::::::::::::::::::::;%#!?#<::::$%?%%@>>@?#@:.    
		 .#%::::::&!;;;@;;>>:>?!!:::;?><@::::::::;&%#<:::>??%;:::::>##>;;;;!#<::#%???$!;#B&_.     
		 .<&;::::;B;;;;@:::::#;:<%?!<!%#<:::::::::&!<;:<#?:>&?::::>B&%!;;;>!?@!:%#???&&>$!.       
		  ,;?%###$@??!!$!>>;;::::::;>;:::::::::::::!%%%!;::;;::::::::&%;;;?#;:::?&????B@;.        
		     ...,_-~=++:;;><<!???%%%###%%??!!<<>>>;;:::::::::::::::::@>;;;%!::::!&??%$&~.         
		                        ........,__-~~==+::;;>>>><<<<!!!!???#@%###@&##%%#&%?>~..          
		Mutable chain implementation of mapped container with objects of any type.
		It is a special dict, as all logic lies behind maintaining "_inside_" dict and "_locker_" list,
		containing only keys, to be a part of mutable chain.
		One can put whatever key-value pair into KeyChest by calling the KeyChest, providing
		key-value pair as arguments. All arguments after key-value pair will be considered as nesting keys,
		so by the special flag "mapped" their will either be created or not, with warnings logs.
		Providing no arguments during KeyChest call will simply return a copy of "_inside_" dict
		(calling with arguments to put in KeyChest will also return such copy).
		KeyChest can be iterated, compared to be equal to another KeyChest or a dict, compared to be
		greater (or equal)/lesser (or equal) than another KeyChest or dict (comparison will take place
		against lengths), key-value pairs can be assigned/obtained/deleted. The built-in function
		"next" will take place instantly, so KeyChest is iterator by default. There is no "full exhaustion"
		of such iterator, so after StopIteration raised when all "_inside_" elements passed, the next
		call of "next" (after call that raised StopIteration), the iteration starts over again.
		It is prohibited to modify "_inside_" and "_locker_" directly, cause it might brake KeyChest!
	"""

	def __len__(self) -> int : return len(self._inside_)
	def __hash__(self) -> int : return id(self)
	def __getitem__(self, K :Hashable) -> Any | None :

		"""
			Despite python core logic implies TypeError raise when unhashable key is accessed
			even with "get" method, KeyChest "__getitem__" implementation return None in such
			situation, because pygwarts is about magic.
		"""

		try:	return		self._inside_.get(K)
		except	TypeError : return




	def __contains__(self, K :Hashable)	-> bool	:

		"""
			Despite python core logic implies TypeError raise when unhashable key is even peeked
			for presence in dictionary, KeyChest "__contains__" implementation return False in such
			situation, because pygwarts is about magic.
		"""

		try:	return		K in self._inside_
		except	TypeError : return	False




	def __setitem__(self, K :Hashable, V :Any):

		"""
			Assigning "_inside_" dict key-value pair, along with maintenance of "_locker_" list of keys.
			If key is a new one, both "_inside_" and "_locker_" will be updated. If key is already present,
			only "_inside_" will be modified. If key is not hashable type, TypeError will be caught and
			warning message will be logged, with no KeyChest object modification.
		"""

		try:

			presence = K in self._inside_
			self._inside_[K] = V


			if	presence : self.loggy.debug(f"Replaced \"{K}\", \"{V}\" pair")
			else:

				self._locker_.append(K)
				self.loggy.debug(f"Putted \"{K}\", \"{V}\" pair")
		except	TypeError : self.loggy.warning(f"Can't put \"{K}\", \"{V}\" pair in {self} KeyChest")




	def __delitem__(self, K :Hashable):

		"""
			Removing key from "_inside_" dict and "_locker_" list. If key is not present in KeyChest,
			corresponding debug message will be logged. If key is not hashable type, TypeError will be
			caught and corresponding warning message will be logged.
		"""

		try:

			if	K in self._inside_:

				V = self._inside_[K]
				I = self._locker_.index(K)


				del self._inside_[K]
				del self._locker_[I]


				self.loggy.debug(f"Removed \"{K}\", \"{V}\" pair")
			else:
				self.loggy.debug(f"Key \"{K}\" not in {self} KeyChest")
		except	TypeError :	self.loggy.warning(f"Impossible key \"{K}\" for {self} KeyChest")
		except	ValueError: self.loggy.warning(f"Key \"{K}\" not found in locker")




	def __eq__(self, other :KeyChest | dict) -> bool :

		"""
			Comparison of equality of current KeyChest "_inside_" dict with another KeyChest
			object's "_inside_" dict or with any other dict object.
		"""

		match other:

			case KeyChest()	: return	self._inside_ == other._inside_
			case dict()		: return	self._inside_ == other
			case _			: return	False




	def __gt__(self, other :KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of greatness of current KeyChest "_inside_" dict length with another KeyChest
			object's "_inside_" dict or with any other dict object's length.
		"""

		match other:

			case KeyChest()	: return	len(self._inside_) > len(other._inside_)
			case dict()		: return	len(self._inside_) > len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} KeyChest")




	def __ge__(self, other :KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of greatness or equality of current KeyChest "_inside_" dict length with another
			KeyChest object's "_inside_" dict or with any other dict object's length.
		"""

		match other:

			case KeyChest()	: return	len(self._inside_) >= len(other._inside_)
			case dict()		: return	len(self._inside_) >= len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} KeyChest")




	def __lt__(self, other :KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of lesserness of current KeyChest "_inside_" dict length with another KeyChest
			object's "_inside_" dict or with any other dict object's length.
		"""

		match other:

			case KeyChest()	: return	len(self._inside_) < len(other._inside_)
			case dict()		: return	len(self._inside_) < len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} KeyChest")




	def __le__(self, other :KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of lesserness or equality of current KeyChest "_inside_" dict length with another
			KeyChest object's "_inside_" dict or with any other dict object's length.
		"""

		match other:

			case KeyChest()	: return	len(self._inside_) <= len(other._inside_)
			case dict()		: return	len(self._inside_) <= len(other)
			case _			: raise		TypeError(f"Object \"{other}\" cannot be compared with {self} KeyChest")




	def __iter__(self):

		"""
			Iterator implementation. Iteration occurs along "_locker_" list, which contain only keys
			to index into "_inside_" dict. Such logic implements the ordered by injection iteration
			by default in any situation. Yields a tuple of both key and value on every iteration.
			Logs a debug message before every yield. If there is no keys in "_locker_", logs
			corresponding debug message.
		"""

		if	len(self._locker_):
			for K in self._locker_:

				V = self._inside_[K]
				self.loggy.debug(f"Yielding \"{K}\", \"{V}\" pair")


				yield K,V
		else:	self.loggy.debug(f"KeyChest is empty")




	def __next__(self):

		"""
			KeyChest the iterator by default, so calling next will iterate through "_locker_" keys,
			yielding a tuple of both key and value. Apart to generator like iteration,
			KeyChest iteration will be reset every time StopIteration is raised after "_locker_" exhausted,
			so it might be iterated over and over again. To make "next" calls work classic way,
			it is possible to call it over "iter" call on current KeyChest.
		"""

		if	self._current < len(self._locker_):

			_next_key_ = self._locker_[self._current]
			self._current += 1


			return	_next_key_, self._inside_[_next_key_]


		self._current = 0
		raise	StopIteration




	def __reversed__(self):

		""" Reversion along "_locker_" to obtain "_inside_" backward. """

		for key in reversed(self._locker_) : yield key, self._inside_[key]




	def __call__(self, *load, mapped=True) -> dict | None :

		"""
			When load is a two items tuple, they will be assigned as corresponding key value pair
			for "_inside_" dict. If such key is not already present in "_inside_" dict, it will be
			added to the "_locker_" list, that maintains only keys of current KeyChest.
			Any arguments after key-value pair will be considered as nesting keys, which will be
			tried to be created, in case of their absent and "mapped" toggled to False, or indexed
			to put key-value pair at the very end, with corresponding warning loggs in case of any
			trouble and None as the returned value.
			An empty load is simply "_inside_" copy returning, cause __str__ will never return
			"_inside_", so such call with no arguments is the way to obtain "_inside_" dict copy.
			Any other situation will cause no KeyChest modification, corresponding warning message
			and None as the returned value.
		"""

		match load:

			# Regular put. Key-value pair (K,V) will be either putted or replaced, depending on
			# whether current KeyChest already contain such key or not.
			case ( K,V ):
				try:

					presence = K in self._inside_
					self._inside_[K] = V


					# Maintaining "_locker_" list, that tracks root keys in order their appears.
					if	presence : self.loggy.debug(f"Replaced \"{K}\", \"{V}\" pair")
					else:

						self._locker_.append(K)
						self.loggy.debug(f"Putted \"{K}\", \"{V}\" pair")


				# Must be unhashable key raise.
				except	TypeError:

					self.loggy.warning(f"Failed to put unhashable key \"{K}\" in {self} KeyChest")
					return




			# Nested put. Key-value pair (K,V) will be putted to nested dictionary, that must be found
			# by following through the dictionary, that must be nested in "_inside_", with keys, that
			# packed in "nest". Boolean "mapped" value is opposite to whether keys from "nest" pack
			# might be created or not (this boolean value means if such keys must present or not).
			case ( K,V, *nest ):

				nesting		= self._inside_
				nest_root	= nest[0]


				try:

					presence = nest_root in nesting
					self.loggy.debug(f"Root key \"{nest_root}\" nesting")


				except	TypeError:

					self.loggy.warning(f"Unhashable root key \"{nest_root}\" for {self} KeyChest")
					return




				for key in nest:
					try:

						if	(next_nesting := nesting.get(key)) is not None:

							self.loggy.debug(f"Nesting key \"{key}\"")
							nesting = next_nesting


						# Nesting KeyChest with dictionaries.
						elif not mapped:

							nesting[key] = dict()
							nesting = nesting[key]
							self.loggy.debug(f"Mapped nested key \"{key}\"")


						else:
							self.loggy.warning(f"Can't nest unmapped key \"{key}\" in {self} KeyChest")
							return


					# Must be unhashable key raise.
					except	TypeError:

						self.loggy.warning(f"Failed to nest unhashable key \"{key}\" in {self} KeyChest")
						return

					# nesting value must be dict or implement get method.
					except	AttributeError:

						self.loggy.warning(f"Unnestable value \"{nesting}\" in {self} KeyChest")
						return




				if	not isinstance(nesting, dict):

					self.loggy.warning(f"Value \"{nesting}\" cannot be nested in {self} KeyChest")
					return


				try:

					# NP - Nested Presence.
					NP = K in nesting
					nesting[K] = V


					if	NP:	self.loggy.debug(f"Replaced nested \"{K}\", \"{V}\" pair")
					else:	self.loggy.debug(f"Putted nested \"{K}\", \"{V}\" pair")


				# Must be unhashable key raise.
				except	TypeError:

					self.loggy.warning(f"Failed to nest unhashable key \"{K}\" in {self} KeyChest")
					return




				# Maintaining "_locker_" list, that tracks root keys in order their appears.
				if	presence: self.loggy.debug(f"Changed root key \"{nest_root}\" nested values")
				else:

					self._locker_.append(nest_root)
					self.loggy.debug(f"Putted nested root key \"{nest_root}\"")




			case ()	: pass
			case _	:

				self.loggy.warning(f"Can't put \"{load}\" in {self} KeyChest")
				return


		# The returned value is intentionally set up to be a copy of an "_inside_" dict, with
		# keys set to be strings. The purpose of such modification is to maintain Transmutable
		# keys to be easy to read, as such dictionary, that __call__ returning, is an another way
		# just to inspect KeyChest, not to interact with it.
		return	{ str(K):V for K,V in self._inside_.items() }




	def __init__(self, *args, **kwargs):

		# Initiating KeyChest core fields first allows putting during initiation.
		self.unload()
		super().__init__(*args, **kwargs)




	def deep(self, *depth) -> Any | None :

		"""
			Deep keying into a KeyChest nestings.
			First argument will be considered as KeyChest root key, all consequent arguments are nested
			keys to go deep in KeyChest. Return nested value or None if neither root key nor any of nested
			keys doesn't exist in KeyChest.
		"""

		if	depth:

			root, *rest = depth
			self.loggy.debug(f"Diving depth {len(rest)}")


			if	(bottom := self._inside_.get(root)) is not None:


				for deep_key in rest:
					if	(bottom := bottom.get(deep_key)) is not None : continue


					self.loggy.debug(f"Deep-key \"{deep_key}\" not found")
					break
				else:
					self.loggy.debug(f"Deep-value found: \"{bottom}\"")
					return	bottom
			else:	self.loggy.debug(f"Root-key \"{root}\" not found")
		else:		self.loggy.debug(f"No depth to go deep")




	def keys(self) -> List[Hashable]:

		"""
			Classic dictionary method implementation. Returns "_locker_" list copy (as new list).
			This method along with "__getitem__" must ensure KeyChest object will be recognized
			to implement mapping protocol.
		"""

		return	list(self._locker_)




	def keysof(self, value :Any) -> List[Hashable] | List :

		"""
			Finding list of keys, that are matched with "value". The returned list of keys is ordered by
			key injection, as it is obtained from "_locker_", just like the iteration. If no match found,
			empty list will be returned and corresponding debug message will be logged.
		"""

		if	value in self._inside_.values():

			self.loggy.debug(f"Matched keys for \"{value}\" are found")
			return	[ K for K in self._locker_ if self._inside_.get(K) == value ]


		self.loggy.debug(f"No matched keys for \"{value}\" are found")
		return	list()




	def unload(self):

		"""
			Wipes "_inside_" and "_locker_" items completely by reinitiating with "_current".
			Also used as initiation by itself.
		"""

		self._inside_ = dict()
		self._locker_ = list()
		self._current = 0








	class Inspection(Transmutable):

		"""
			Producer of an elegant and easy to read string of current KeyChest content.
			The produced string is kind of "pretty print" implemented.
			Also allow logging content in any level, that must be provided (no logging by default).
		"""

		def __call__(self, level :int =None) -> str :

			self.loggy.info(f"Commencing {self._UPPER_LAYER} inspection")
			self.view = str()


			self.inspecting(log=level)
			self.loggy.debug(f"Inspection result {len(self.view)} symbols")


			return	self.view




		def inspecting(self, log, layer :dict =None, deep :int =0):

			for k,v in (_layer := layer.items() if layer else self._UPPER_LAYER) :

				_k = deep *"\t" +str(k)
				self.view += f"{_k}:\n"
				if	log : self.loggy.log(log, _k)


				if	v and isinstance(v, dict) : self.inspecting(log, v, deep +1)
				else:

					_v = (deep +1) *"\t" +str(v)
					self.view += f"{_v}\n"
					if	log : self.loggy.log(log, _v)










import	os
import	unittest
from	time								import sleep
from	pygwarts.tests.hagrid				import EasySet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.shelve				import LibraryShelf
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.bloom.twigs			import Germination
from	pygwarts.hagrid.bloom.leafs			import Rejuvenation
from	pygwarts.hagrid.planting.leafs		import LeafGrowth
from	pygwarts.hagrid.planting.twigs		import TwigThrive
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.planting.peeks		import BlindPeek
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class BoughArgumentCase(EasySet):

	"""
		GrowingPeel, BlindPeek, LeafGrowth, Germination -> Rejuvenation
		Single sprout
		Single bough (passed as argument in init)
		This test is duplication for easy1, except bough providing
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.EASY_HANDLER_13): os.remove(cls.EASY_HANDLER_13)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.EASY_HANDLER_13)

	def setUp(self):
		class Sakura(Tree):
			class loggy(LibraryContrib):

				handler		= self.EASY_HANDLER_13
				init_level	= 10

			@GrowingPeel
			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			@GrowingPeel
			@BlindPeek("seeds", renew=False)
			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			class seeds(LibraryShelf):

				grabbing	= self.EASY_SET_SEEDS
				reclaiming	= self.EASY_SET_SEEDS

			@fssprout(self.EASY_SET_SPROUT)
			class sync(Flourish):

				class twigs(SiftingController):

					include	= (

						( os.path.join(self.EASY_SET_SPROUT, "pros"), )
						if os.name == "posix" else
						( os.path.join(self.EASY_SET_SPROUT, "pros").replace("\\", "\\\\"), )
					)
					exclude	= (

						( os.path.join(self.EASY_SET_SPROUT, "pros", ".+"), )
						if os.name == "posix" else
						( os.path.join(self.EASY_SET_SPROUT, "pros", ".+").replace("\\", "\\\\"), )
					)

				class leafs(SiftingController):

					include	= r".+\.txt$",
					exclude	= (

						( rf".+/[^/]+good.*\.txt$", )
						if os.name == "posix" else
						( rf".+\\[^(\\)]+good.*\.txt$", )
					)

		self.test_case = Sakura




	def test_first_flourish(self):

		self.test_case.loggy.init_name = "first_flourish"
		self.assertTrue(os.path.isdir(self.EASY_SET_BOUGH))
		with self.assertLogs("first_flourish", 10) as case_loggy:

			self.test_case = self.test_case(bough=self.EASY_SET_BOUGH)
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		# Included file "WitchDoctor.txt"
		self.assertTrue(os.path.isfile(self.dst_file1))
		# Included folder "pros"
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		# File excluded by ".+{os.sep}[^{os.sep}]+good.*\.txt$"
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		# Files included by ".+\.txt$"
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		# Folders and files sifted out by being not included
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))

		self.assertIn(f"INFO:first_flourish:Grown leaf \"{self.dst_file1}\"", case_loggy.output)
		self.assertIn(f"INFO:first_flourish:Thrived twig \"{self.dst_pros_folder}\"", case_loggy.output)
		self.assertIn(f"INFO:first_flourish:Grown leaf \"{self.dst_pros_file2}\"", case_loggy.output)
		self.assertIn(f"INFO:first_flourish:Grown leaf \"{self.dst_pros_file3}\"", case_loggy.output)
		self.assertIn(f"DEBUG:first_flourish:Sifted sprig \"cons\"", case_loggy.output)
		self.assertIn(f"DEBUG:first_flourish:Sifted sprig \"not so good.txt\"", case_loggy.output)
		self.assertIn(f"DEBUG:first_flourish:Sifted sprig \"redundant_folder_1\"", case_loggy.output)
		self.assertIn(f"DEBUG:first_flourish:Sifted sprig \"redundant_folder_2\"", case_loggy.output)
		self.assertNotIn(f"INFO:first_flourish:Grown leaf \"{self.dst_pros_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:first_flourish:Grown leaf \"{self.dst_cons_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:first_flourish:Grown leaf \"{self.dst_cons_file2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:first_flourish:Grown leaf \"{self.dst_redundant_1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:first_flourish:Grown leaf \"{self.dst_redundant_2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:first_flourish:Thrived twig \"{self.dst_cons_folder}\"", case_loggy.output)
		self.assertNotIn(

			f"INFO:first_flourish:Thrived twig \"{self.dst_redundant_1_folder}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:first_flourish:Thrived twig \"{self.dst_redundant_2_folder}\"",
			case_loggy.output
		)

		self.assertEqual(len(self.test_case.seeds),0)
		self.assertEqual(len(self.test_case.seeds()),3)
		self.assertNotIn(self.file1, self.test_case.seeds)
		self.assertNotIn(self.pros_file2, self.test_case.seeds)
		self.assertNotIn(self.pros_file3, self.test_case.seeds)
		self.assertIn(self.file1, self.test_case.seeds())
		self.assertIn(self.pros_file2, self.test_case.seeds())
		self.assertIn(self.pros_file3, self.test_case.seeds())
		self.assertFalse(

			os.path.isfile(self.EASY_SET_SEEDS)
			or
			os.path.isfile(self.EASY_SET_SEEDS + ".db")
			or
			os.path.isfile(self.EASY_SET_SEEDS + ".dat")
		)
		self.test_case.seeds.produce(magical=True, rewrite=True)
		self.assertTrue(

			os.path.isfile(self.EASY_SET_SEEDS)
			or
			os.path.isfile(self.EASY_SET_SEEDS + ".db")
			or
			os.path.isfile(self.EASY_SET_SEEDS + ".dat")
		)








	def test_no_touch_flourish(self):

		sleep(1.1)
		self.test_case.loggy.init_name = "no_touch_flourish"
		with self.assertLogs("no_touch_flourish", 10) as case_loggy:

			self.test_case = self.test_case(bough=self.EASY_SET_BOUGH)
			self.test_case.sync()
			self.test_case.seeds.produce(magical=True, rewrite=True)


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertIn("DEBUG:no_touch_flourish:Shelf was not modified", case_loggy.output)
		self.assertIn("DEBUG:no_touch_flourish:Flourish stopped by blind peek comparator", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Thrived twig \"{self.dst_pros_folder}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_pros_file2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_pros_file3}\"", case_loggy.output)
		self.assertIn(f"DEBUG:no_touch_flourish:Sifted sprig \"cons\"", case_loggy.output)
		self.assertIn(f"DEBUG:no_touch_flourish:Sifted sprig \"not so good.txt\"", case_loggy.output)
		self.assertIn(f"DEBUG:no_touch_flourish:Sifted sprig \"redundant_folder_1\"", case_loggy.output)
		self.assertIn(f"DEBUG:no_touch_flourish:Sifted sprig \"redundant_folder_2\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_pros_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_cons_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_cons_file2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_redundant_1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Grown leaf \"{self.dst_redundant_2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:no_touch_flourish:Thrived twig \"{self.dst_cons_folder}\"", case_loggy.output)
		self.assertNotIn(

			f"INFO:no_touch_flourish:Thrived twig \"{self.dst_redundant_1_folder}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:no_touch_flourish:Thrived twig \"{self.dst_redundant_2_folder}\"",
			case_loggy.output
		)








	def test_touch_flourish(self):

		sleep(1.1)
		self.test_case.loggy.init_name = "touch_flourish"
		self.fmake(self.pros_file3, "definitely the best way")
		self.fmake(self.cons_file1, "might cause a headache")

		with self.assertLogs("touch_flourish", 10) as case_loggy:

			self.test_case = self.test_case(bough=self.EASY_SET_BOUGH)
			self.test_case.sync()


		self.no_loggy_levels(case_loggy.output, 30,40,50)


		# Included file "WitchDoctor.txt"
		self.assertTrue(os.path.isfile(self.dst_file1))
		# Included folder "pros"
		self.assertTrue(os.path.isdir(self.dst_pros_folder))
		# File excluded by ".+{os.sep}[^{os.sep}]+good.*\.txt$"
		self.assertFalse(os.path.isfile(self.dst_pros_file1))
		# Files included by ".+\.txt$"
		self.assertTrue(os.path.isfile(self.dst_pros_file2))
		self.assertTrue(os.path.isfile(self.dst_pros_file3))
		# Folders and files sifted out by being not included
		self.assertFalse(os.path.isdir(self.dst_cons_folder))
		self.assertFalse(os.path.isfile(self.dst_cons_file1))
		self.assertFalse(os.path.isfile(self.dst_cons_file2))
		self.assertFalse(os.path.isdir(self.dst_redundant_1_folder))
		self.assertFalse(os.path.isdir(self.dst_redundant_2_folder))
		self.assertFalse(os.path.isfile(self.dst_redundant_1))
		self.assertFalse(os.path.isfile(self.dst_redundant_2))

		self.assertIn("DEBUG:touch_flourish:Flourish stopped by blind peek comparator", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Thrived twig \"{self.dst_pros_folder}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_pros_file2}\"", case_loggy.output)
		self.assertIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_pros_file3}\"", case_loggy.output)
		self.assertIn(f"DEBUG:touch_flourish:Sifted sprig \"cons\"", case_loggy.output)
		self.assertIn(f"DEBUG:touch_flourish:Sifted sprig \"not so good.txt\"", case_loggy.output)
		self.assertIn(f"DEBUG:touch_flourish:Sifted sprig \"redundant_folder_1\"", case_loggy.output)
		self.assertIn(f"DEBUG:touch_flourish:Sifted sprig \"redundant_folder_2\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_pros_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_cons_file1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_cons_file2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_redundant_1}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Grown leaf \"{self.dst_redundant_2}\"", case_loggy.output)
		self.assertNotIn(f"INFO:touch_flourish:Thrived twig \"{self.dst_cons_folder}\"", case_loggy.output)
		self.assertNotIn(

			f"INFO:touch_flourish:Thrived twig \"{self.dst_redundant_1_folder}\"",
			case_loggy.output
		)
		self.assertNotIn(

			f"INFO:touch_flourish:Thrived twig \"{self.dst_redundant_2_folder}\"",
			case_loggy.output
		)

		self.assertEqual(len(self.test_case.seeds),3)
		self.assertEqual(len(self.test_case.seeds()),3)
		self.assertIn(self.file1, self.test_case.seeds)
		self.assertIn(self.pros_file2, self.test_case.seeds)
		self.assertIn(self.pros_file3, self.test_case.seeds)
		self.assertIn(self.file1, self.test_case.seeds())
		self.assertIn(self.pros_file2, self.test_case.seeds())
		self.assertIn(self.pros_file3, self.test_case.seeds())
		self.test_case.seeds.produce(magical=True, rewrite=True)
		self.assertTrue(

			os.path.isfile(self.EASY_SET_SEEDS)
			or
			os.path.isfile(self.EASY_SET_SEEDS + ".db")
			or
			os.path.isfile(self.EASY_SET_SEEDS + ".dat")
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing							import Optional
from typing							import Callable
from pygwarts.irma.access.inducers	import AccessInducer
from pygwarts.irma.access.volume	import LibraryVolume








class HandlerCounterInducer(AccessInducer):

	"""
		Implementation of inducer, that accesses the AccessHandler object it is assigned to and returns
		it's "access_handler_counter" attribute, which must be integer that represents the counter for
		AccessHandler invocations. Processes obtained value with special field "filter", which must be
		callable that accepts two arguments - current AccessInducer object and the value it processing.
		It is must be noted, that "filter" invocation implemented according to "filter" assigned during
		declaration of AccessInducer class, so it is instantiated. If "filter" will be assigned to already
		instantiated AccessInducer object, it must account for accepting the value argument only. It doesn't
		restricted for "filter" what value to return, cause it anyway will be converted to string. Returns
		either obtained, and perhaps processed, value if it is originally not None or "filter" returned
		not None. Otherwise returns None. There is no Exception handling for "filter" invocation.
	"""

	filter :Optional[Callable[[AccessInducer,int],int | None]]

	def __call__(self, volume :LibraryVolume) -> str | None :
		if	isinstance((counter := self.get_handler_counter()), int):

			if		callable(getattr(self, "filter", None)): counter = self.filter(counter)
			else:	self.loggy.debug("Filter not applied")
			if		counter is not None : return str(counter)








class RegisterCounterInducer(AccessInducer):

	"""
		Implementation of inducer, that accesses provided "volume" argument LibraryVolume object with
		AccessHandler object it is assigned to, so it must be it's "_UPPER_LAYER", and returns
		corresponding mapped "counter" value, which must be integer that represents the counter of
		AccessHandler invocations for exactly "volume" object. Processes obtained value with special
		field "filter", which must be callable that accepts two arguments - current AccessInducer object
		and the value it processing. It is must be noted, that "filter" invocation implemented according
		to "filter" assigned during declaration of AccessInducer class, so it is instantiated. If "filter"
		will be assigned to already instantiated AccessInducer object, it must account for accepting the
		value argument only. It doesn't restricted for "filter" what value to return, cause it anyway will
		be converted to string. Returns either obtained, and perhaps processed, value if it is originally
		not None or "filter" returned not None. Otherwise returns None. There is no Exception handling for
		"filter" invocation.
	"""

	filter :Optional[Callable[[AccessInducer,int],int | None]]

	def __call__(self, volume :LibraryVolume) -> str | None :
		if	isinstance((counter := self.get_register_counter(volume)), int):

			if		callable(getattr(self, "filter", None)): counter = self.filter(counter)
			else:	self.loggy.debug("Filter not applied")
			if		counter is not None : return str(counter)







